from collections import defaultdict
from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    TYPE_CHECKING,
    TypeVar,
)
from grams.algorithm.data_graph.dg_graph import DGGraph
from grams.algorithm.inferences.features.tree_utils import TreeStruct
from grams.algorithm.inferences.psl_lib import IDMap
from sm.misc import CacheMethod
import numpy as np
from grams.algorithm.data_graph import CellNode
from grams.algorithm.literal_matchers import TextParser
from grams.algorithm.candidate_graph.cg_graph import CGColumnNode, CGGraph
from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import (
    WDEntity,
    WDProperty,
    WDQuantityPropertyStats,
    WDClass,
)
from operator import itemgetter


class TypeFeaturesConfigs:
    EXTENDED_DISTANCE: int = 1


class TypeFeatures:
    INSTANCE_OF = "P31"
    HIERARCHY_PROPS = {"P131", "P276"}

    def __init__(
        self,
        idmap: IDMap,
        table: LinkedTable,
        cg: CGGraph,
        dg: DGGraph,
        wdentities: Mapping[str, WDEntity],
        wdclasses: Mapping[str, WDClass],
        wdprops: Mapping[str, WDProperty],
        wd_num_prop_stats: Mapping[str, WDQuantityPropertyStats],
        sim_fn: Optional[Callable[[str, str], float]] = None,
    ):
        self.idmap = idmap
        self.table = table
        self.cg = cg
        self.dg = dg
        self.wdentities = wdentities
        self.wdclasses = wdclasses
        self.wdprops = wdprops
        self.wd_num_prop_stats = wd_num_prop_stats
        self.sim_fn = sim_fn

    def extract_features(self, features: List[str]) -> Dict[str, list]:
        # gather the list of entity columns (columns will be tagged with type)
        tagged_columns = []
        for u in self.cg.iter_nodes():
            if not isinstance(u, CGColumnNode):
                continue

            cells = self.get_cells(u)

            # using heuristic to determine if we should tag this column
            covered_fractions = [
                sum(
                    span.length
                    for spans in cell.entity_spans.values()
                    for span in spans
                )
                / max(len(cell.value), 1)
                for cell in cells
                if len(cell.entity_ids) > 0
            ]

            if len(covered_fractions) == 0:
                continue

            avg_covered_fractions = np.mean(covered_fractions)
            if avg_covered_fractions < 0.8:
                if avg_covered_fractions == 0:
                    avg_cell_len = np.mean(
                        [len(cell.value) for cell in cells if len(cell.entity_ids) > 0]
                    )
                    if avg_cell_len < 1:
                        # links are likely to be image such as national flag, so we still model them
                        tagged_columns.append(u)
                continue

            tagged_columns.append(u)

        # extract features
        feat_data = {}
        for feat in features:
            fn = getattr(self, feat)
            feat_data[feat] = [v for u in tagged_columns for v in fn(u)]
        return feat_data

    def TYPE_FREQ_OVER_ROW(self, u: CGColumnNode) -> List[Tuple[str, str, float]]:
        type_freq = self.get_type_freq(u)
        output = []
        n_rows = self.table.size()
        for c, freq in type_freq.items():
            output.append((self.idmap.m(u.id), self.idmap.m(c), freq / n_rows))
        return output

    def EXTENDED_TYPE_FREQ_OVER_ROW(
        self, u: CGColumnNode
    ) -> List[Tuple[str, str, float]]:
        extended_type_freq = self.get_extended_type_freq(u)
        output = []
        n_rows = self.table.size()
        for c, freq in extended_type_freq.items():
            output.append((self.idmap.m(u.id), self.idmap.m(c), freq / n_rows))
        return output

    def TYPE_FREQ_OVER_ENT_ROW(self, u: CGColumnNode) -> List[Tuple[str, str, float]]:
        type_freq = self.get_type_freq(u)
        output = []
        n_ent_rows = self.get_num_ent_rows(u)
        for c, freq in type_freq.items():
            output.append((self.idmap.m(u.id), self.idmap.m(c), freq / n_ent_rows))
        return output

    def EXTENDED_TYPE_FREQ_OVER_ENT_ROW(
        self, u: CGColumnNode
    ) -> List[Tuple[str, str, float]]:
        extended_type_freq = self.get_extended_type_freq(u)
        output = []
        n_ent_rows = self.get_num_ent_rows(u)
        for c, freq in extended_type_freq.items():
            output.append((self.idmap.m(u.id), self.idmap.m(c), freq / n_ent_rows))
        return output

    def HAS_SUB_TYPE(self, u: CGColumnNode):
        output = set()
        classes = self.get_extended_type_freq(u)
        for c in classes:
            for pc in self.wdclasses[c].parents:
                if pc in classes:
                    output.add((self.idmap.m(u.id), self.idmap.m(pc)))
        return list(output)

    def ABS_TYPE_DISTANCE(self, u: CGColumnNode):
        """Get types and distance from the class to its most specific type."""
        output = set()
        classes = self.get_extended_type_freq(u)
        tree = self.get_candidate_types_as_maps(u)

        # start with leaf nodes
        start_nodes: List[Tuple[str, int]] = [
            (c, 0) for c in classes if len(tree.p2cs[c]) == 0
        ]
        best_distance = {}
        while len(start_nodes) > 0:
            node, distance = start_nodes.pop()
            best_distance[node] = min(distance, best_distance.get(node, float("inf")))
            for child in tree.c2ps[node]:
                start_nodes.append((child, distance + 1))

        # longest distance, can't use distance to root because root node
        # can even have shorter distance than one of its children due to varied tree length
        longest_distance = max(1, max(best_distance.values()))
        for c, distance in best_distance.items():
            output.add(
                (self.idmap.m(u.id), self.idmap.m(c), distance / longest_distance)
            )
        return list(output)

    def TYPE_DISTANCE(self, u: CGColumnNode):
        output = set()
        extended_classes = self.get_extended_type_freq(u)
        if len(extended_classes) == 0:
            return []

        classes = self.get_type_freq(u)
        tree = self.get_candidate_types_as_maps(u)

        top_classes = sorted(classes.items(), key=itemgetter(1), reverse=True)[:2]
        # now calculate the distance from the top classes into its parents
        base = set(c for c, f in top_classes)
        best_distance: Dict[str, float] = {c: 0 for c in base}
        for c in base:
            stack: List[Tuple[str, float]] = [
                (p, 1) for p in tree.c2ps[c] if p not in base
            ]
            while len(stack) > 0:
                node, distance = stack.pop()
                best_distance[node] = min(
                    distance, best_distance.get(node, float("inf"))
                )
                for p in tree.c2ps[node]:
                    if p not in base:
                        stack.append((p, distance + 1))

        longest_distance = max(1, max(best_distance.values()))

        for c in extended_classes:
            if c not in best_distance:
                # we can't reach these classes from the top classes
                # TODO: what should be the value? for now, we just set it to be half of the max value (50%)
                best_distance[c] = longest_distance / 2

        for c, distance in best_distance.items():
            output.add(
                (self.idmap.m(u.id), self.idmap.m(c), distance / longest_distance)
            )
        return list(output)

    def TYPE_HEADER_SIMILARITY(self, u: CGColumnNode):
        if self.sim_fn is None:
            return []

        extended_type_freq = self.get_extended_type_freq(u)
        output = []

        for c, freq in extended_type_freq.items():
            sim = self.sim_fn(
                u.label,
                self.wdclasses[c].label,
            )
            output.append((self.idmap.m(u.id), self.idmap.m(c), sim))
        return output

    @CacheMethod.cache(CacheMethod.single_object_arg)
    def get_type_freq(self, u: CGColumnNode) -> Dict[str, float]:
        """Calculating frequency of types in a column.
        Each time a type appears in a cell, instead of counting 1, we count its probability
        """
        cells = self.get_cells(u)
        cell2qnodes = self.get_cell_to_qnodes(u)

        type2freq = {}

        for cell in cells:
            classes = {}
            for qnode, prob in cell2qnodes[cell.id]:
                for stmt in qnode.props.get(self.INSTANCE_OF, []):
                    assert stmt.value.is_entity_id(stmt.value), stmt.value
                    stmt_value_ent_id = stmt.value.as_entity_id()
                    if stmt_value_ent_id not in self.wdclasses:
                        # sometimes people just tag things incorrectly, e.g.,
                        # Q395 is not instanceof Q41511 (Q41511 is not a class)
                        continue
                    classes[stmt_value_ent_id] = max(
                        prob, classes.get(stmt_value_ent_id, 0)
                    )

            for c, prob in classes.items():
                if c not in type2freq:
                    type2freq[c] = 0.0
                type2freq[c] += prob

        return type2freq

    @CacheMethod.cache(CacheMethod.single_object_arg)
    def get_extended_type_freq(self, u: CGColumnNode) -> Dict[str, float]:
        """Calculating frequency of types in a column.
        Each time a type appears in a cell, instead of counting 1, we count its probability
        """
        cells = self.get_cells(u)
        cell2qnodes = self.get_cell_to_qnodes(u)

        extended_type2freq = {}

        for cell in cells:
            classes = {}
            for qnode, prob in cell2qnodes[cell.id]:
                for stmt in qnode.props.get(self.INSTANCE_OF, []):
                    stmt_value_ent_id = stmt.value.as_entity_id_safe()
                    if stmt_value_ent_id not in self.wdclasses:
                        # sometimes people just tag things incorrectly, e.g.,
                        # Q395 is not instanceof Q41511 (Q41511 is not a class)
                        continue
                    classes[stmt_value_ent_id] = max(
                        prob, classes.get(stmt_value_ent_id, 0)
                    )
            classes = self._get_extended_type_freq_of_cell(
                classes, TypeFeaturesConfigs.EXTENDED_DISTANCE
            )
            for c, prob in classes.items():
                if c not in extended_type2freq:
                    extended_type2freq[c] = 0.0
                extended_type2freq[c] += prob

        return extended_type2freq

    def _get_extended_type_freq_of_cell(self, classes: Dict[str, float], distance: int):
        """Get parents of classes within distance from the node.
        Distance 1 is the parent, Distance 2 is the grandparent.
        Return the classes and their parents
        """
        if distance <= 0:
            return classes
        output = classes.copy()
        for klass, prob in classes.items():
            for parent_klass in self.wdclasses[klass].parents:
                output[parent_klass] = max(output.get(parent_klass, 0), prob)
        if distance > 1:
            return self._get_extended_type_freq_of_cell(output, distance - 1)
        return output

    @CacheMethod.cache(CacheMethod.single_object_arg)
    def get_candidate_types_as_maps(self, u: CGColumnNode) -> TreeStruct:
        """Get candidate types of a column in a form of two maps: one from parent to children, and another one from child to parents
        If there are cycles, we break them
        """
        candidate_types = self.get_extended_type_freq(u)
        return TreeStruct.construct(
            candidate_types, lambda x: self.wdclasses[x].parents
        ).ensure_tree()

    @CacheMethod.cache(CacheMethod.single_object_arg)
    def get_cells(self, u: CGColumnNode) -> List[CellNode]:
        """Get cells of a column node"""
        return [self.dg.get_cell_node(cid) for cid in u.nodes]

    @CacheMethod.cache(CacheMethod.single_object_arg)
    def get_num_ent_rows(self, u: CGColumnNode) -> int:
        """Get number of rows of entities in a column"""
        return sum([1 for cell in self.get_cells(u) if len(cell.entity_ids) > 0])

    @CacheMethod.cache(CacheMethod.single_object_arg)
    def get_cell_to_qnodes(
        self, u: CGColumnNode
    ) -> Dict[str, List[Tuple[WDEntity, float]]]:
        """Get a mapping to associated qnodes of a cell"""
        cells = self.get_cells(u)
        cell2qnodes = {}
        for cell in cells:
            self._add_merge_qnodes(cell, cell2qnodes)
        return cell2qnodes

    def _add_merge_qnodes(
        self, cell: CellNode, cell2qnodes: Dict[str, List[Tuple[WDEntity, float]]]
    ):
        """merge qnodes that are sub of each other
        attempt to merge qnodes (spatial) if they are contained in each other
        we should go even higher order
        """
        assert cell.id not in cell2qnodes
        if len(cell.entity_ids) > 1:
            # attempt to merge qnodes (spatial) if they are contained in each other
            # we should go even higher order
            ignore_qnodes = set()
            for q0_id in cell.entity_ids:
                q0 = self.wdentities[q0_id]
                vals = {
                    stmt.value.as_entity_id_safe()
                    for p in self.HIERARCHY_PROPS
                    for stmt in q0.props.get(p, [])
                }
                for q1_id in cell.entity_ids:
                    if q0_id == q1_id:
                        continue
                    if q1_id in vals:
                        # q0 is inside q1, ignore q1
                        ignore_qnodes.add(q1_id)
            qnode_lst = [
                self.wdentities[q_id]
                for q_id in cell.entity_ids
                if q_id not in ignore_qnodes
            ]
        elif len(cell.entity_ids) > 0:
            qnode_lst = [self.wdentities[cell.entity_ids[0]]]
        else:
            qnode_lst = []

        qnode2prob = {}
        for link in self.table.links[cell.row][cell.column]:
            for c in link.candidates:
                qnode2prob[c.entity_id] = max(
                    qnode2prob.get(c.entity_id, 0), c.probability
                )

        cell2qnodes[cell.id] = [(qnode, qnode2prob[qnode.id]) for qnode in qnode_lst]

    def _print_class_hierarchy(self, classes: List[str]):
        """Print a tree"""
        p2c = {c: [] for c in classes}
        c2p = {c: [] for c in classes}
        for c in classes:
            for p in self.wdclasses[c].parents:
                if p in classes:
                    p2c[p].append(c)
                    c2p[c].append(p)
        roots = [c for c, ps in c2p.items() if len(ps) == 0]
        label = lambda r: f"{self.wdclasses[r].label} ({r})"
        visited = {}
        counter = 0

        for root in roots:
            stack: List[Tuple[str, int]] = [(root, 0)]
            while len(stack) > 0:
                counter += 1
                node, depth = stack.pop()
                postfix = f" (visited at {visited[node]})" if node in visited else ""
                if depth == 0:
                    print(f"{counter}.\t{label(node)}{postfix}")
                else:
                    indent = "│   " * (depth - 1)
                    print(f"{counter}.\t{indent}├── {label(node)}{postfix}")

                if node in visited:
                    continue
                else:
                    visited[node] = counter

                for child in p2c[node]:
                    stack.append((child, depth + 1))



import re
from typing import Tuple, cast
from grams.algorithm.literal_matchers.text_parser import (
    ParsedTextRepr,
    ParsedDatetimeRepr,
)
from grams.algorithm.literal_matchers.types import LiteralMatchKit
from kgdata.wikidata.models import WDValueTime


def time_test(
    kgvalue: WDValueTime, value: ParsedTextRepr, kit: LiteralMatchKit
) -> Tuple[bool, float]:
    """Compare if the value in KG matches with value in the cell

    Args:
        kgvalue: the value in KG
        value: the value in the cell
        kit: objects that may be needed for matching
    Returns:
        a tuple of (whether it's matched, confidence)
    """
    celldt = value.datetime
    if celldt is None:
        return False, 0.0

    kgtime = kgvalue.value
    timestr = kgtime["time"]
    # there are two calendar:
    # gregorian calendar, and julian calendar (just 13 days behind gregorian)
    # https://www.timeanddate.com/calendar/julian-gregorian-switch.html#:~:text=13%20Days%20Behind%20Today,days%20behind%20the%20Gregorian%20calendar.
    # TODO: we need to consider a range for julian calendar
    assert kgtime["calendarmodel"] in {
        "http://www.wikidata.org/entity/Q1985786",
        "http://www.wikidata.org/entity/Q1985727",
    }, kgtime["calendarmodel"]
    # TODO: handle timezone, before/after and precision
    # pass

    # parse timestring
    match = re.match(r"([-+]\d+)-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})Z", timestr)
    assert match is not None
    kgdt = ParsedDatetimeRepr(
        year=int(match.group(1)),
        month=int(match.group(2)),
        day=int(match.group(3)),
        hour=int(match.group(4)),
        minute=int(match.group(5)),
        second=int(match.group(6)),
    )
    if timestr[0] == "-":
        kgdt.year = -kgdt.year  # type: ignore
    else:
        assert timestr[0] == "+"
    for p in ["year", "month", "day", "hour", "minute", "second"]:
        if getattr(kgdt, p) == 0:
            setattr(kgdt, p, None)

    if kgdt == celldt:
        return True, 1.0
    if celldt.has_only_year() and kgdt.year == celldt.year:
        return True, 0.8
    if celldt.first_day_of_year() and kgdt.year == celldt.year:
        return True, 0.75
    return False, 0.0



import os
from dataclasses import dataclass
from operator import itemgetter
from functools import partial
from pathlib import Path
from typing import Dict, MutableMapping, Set, Tuple, Union
from grams.algorithm.candidate_graph.cg_factory import CGFactory
from grams.algorithm.candidate_graph.cg_graph import CGGraph
from grams.algorithm.data_graph.dg_graph import DGGraph
from grams.algorithm.inferences.psl_gram_model import PSLGramModel
from grams.algorithm.inferences.psl_gram_model_exp import PSLGramModelExp
from grams.algorithm.inferences.psl_gram_model_exp2 import PSLGramModelExp2
from grams.algorithm.inferences.psl_lib import PSLModel
from grams.algorithm.literal_matchers import TextParserConfigs, LiteralMatch
from grams.algorithm.postprocessing import (
    MinimumArborescence,
    PairwiseSelection,
    PostProcessingSimplePath,
    SteinerTree,
)
from hugedict.prelude import CacheDict, Parallel
from kgdata.wikidata.models import WDEntityLabel
from loguru import logger
from tqdm import tqdm
import sm.misc as M
import sm.outputs as O
from kgdata.wikidata.db import (
    WDProxyDB,
    get_entity_db,
    get_entity_label_db,
    get_wdclass_db,
    get_wdprop_db,
    get_wdprop_domain_db,
    get_wdprop_range_db,
    query_wikidata_entities,
)
from kgdata.wikidata.models import (
    WDEntity,
    WDClass,
    WDQuantityPropertyStats,
)

import grams.inputs as I
from grams.algorithm.data_graph import DGConfigs, DGFactory
from grams.algorithm.kg_index import KGObjectIndex, TraversalOption

from grams.algorithm.sm_wikidata import WikidataSemanticModelHelper
from grams.config import DEFAULT_CONFIG


@dataclass
class Annotation:
    sm: O.SemanticModel
    # data graph
    dg: DGGraph
    # candidate graph
    cg: CGGraph
    # probabilities of each edge in cg (uid, vid, edge key)
    cg_edge_probs: Dict[Tuple[str, str, str], float]
    # probabilities of types of each column: column index -> type -> probability
    cta_probs: Dict[int, Dict[str, float]]
    # predicted candidate graph where incorrect relations are removed by threshold & post-processing algorithm
    pred_cpa: CGGraph
    # predicted column types
    pred_cta: Dict[int, str]


class GRAMS:
    """Implementation of GRAMS. The main method is `annotate`"""

    def __init__(
        self,
        data_dir: Union[Path, str],
        cfg=None,
        proxy: bool = True,
    ):
        self.timer = M.Timer()
        self.cfg = cfg if cfg is not None else DEFAULT_CONFIG

        with self.timer.watch("init grams db"):
            read_only = not proxy
            self.wdentities = get_entity_db(
                os.path.join(data_dir, "wdentities.db"),
                read_only=read_only,
                proxy=proxy,
            )
            if proxy:
                assert isinstance(self.wdentities, WDProxyDB)
            if os.path.exists(os.path.join(data_dir, "wdentity_labels.db")):
                self.wdentity_labels = get_entity_label_db(
                    os.path.join(data_dir, "wdentity_labels.db"),
                )
            else:
                self.wdentity_labels: MutableMapping[str, WDEntityLabel] = {}
            self.wdclasses = get_wdclass_db(
                os.path.join(data_dir, "wdclasses.db"),
                read_only=read_only,
                proxy=proxy,
            )
            if os.path.exists(os.path.join(data_dir, "wdclasses.fixed.jl")):
                self.wdclasses = self.wdclasses.cache()
                assert isinstance(self.wdclasses, CacheDict)
                for record in M.deserialize_jl(
                    os.path.join(data_dir, "wdclasses.fixed.jl")
                ):
                    cls = WDClass.from_dict(record)
                    self.wdclasses._cache[cls.id] = cls
            self.wdprops = get_wdprop_db(
                os.path.join(data_dir, "wdprops.db"),
                read_only=read_only,
                proxy=proxy,
            )
            if os.path.exists(os.path.join(data_dir, "wdprop_domains.db")):
                self.wdprop_domains = get_wdprop_domain_db(
                    os.path.join(data_dir, "wdprop_domains.db"),
                    read_only=True,
                )
            else:
                self.wdprop_domains = None

            if os.path.exists(os.path.join(data_dir, "wdprop_ranges.db")):
                self.wdprop_ranges = get_wdprop_range_db(
                    os.path.join(data_dir, "wdprop_ranges.db"),
                    read_only=True,
                )
            else:
                self.wdprop_ranges = None

            self.wd_numprop_stats = WDQuantityPropertyStats.from_dir(
                os.path.join(data_dir, "quantity_prop_stats")
            )

        self.update_config()

    def update_config(self):
        """Update the current configuration of the algorithm based on the current configuration stored in this object"""
        for name, value in self.cfg.data_graph.configs.items():
            if not hasattr(DGConfigs, name):
                raise Exception(f"Invalid configuration for data_graph: {name}")
            setattr(DGConfigs, name, value)

        for name, value in self.cfg.literal_matcher.text_parser.items():
            if not hasattr(TextParserConfigs, name):
                raise Exception(
                    f"Invalid configuration for literal_matcher.text_parser: {name}"
                )
            setattr(TextParserConfigs, name, value)

        for name, value in self.cfg.literal_matcher.matchers.items():
            if not hasattr(LiteralMatch, name):
                raise Exception(
                    f"Invalid configuration for literal_matcher.matchers: {name}"
                )
            setattr(LiteralMatch, name, value)

    def annotate(self, table: I.LinkedTable, verbose: bool = False) -> Annotation:
        """Annotate a linked table"""
        wdentity_ids = {
            link.entity_id
            for rlinks in table.links
            for links in rlinks
            for link in links
            if link.entity_id is not None
        }
        wdentity_ids.update(
            (
                candidate.entity_id
                for rlinks in table.links
                for links in rlinks
                for link in links
                for candidate in link.candidates
            )
        )
        if table.context.page_entity_id is not None:
            wdentity_ids.add(table.context.page_entity_id)

        with self.timer.watch("retrieving entities"):
            wdentities = self.get_entities(
                wdentity_ids, n_hop=self.cfg.data_graph.max_n_hop, verbose=verbose
            )
        wdclasses = self.wdclasses.cache()
        wdprops = self.wdprops.cache()

        nonexistent_wdentity_ids = wdentity_ids.difference(wdentities.keys())
        if len(nonexistent_wdentity_ids) > 0:
            logger.info(
                "Removing non-existent entities: {}", list(nonexistent_wdentity_ids)
            )
            table.remove_nonexistent_entities(nonexistent_wdentity_ids)

        with self.timer.watch("retrieving entities label"):
            wdentity_labels = self.get_entity_labels(wdentities, verbose=verbose)

        with self.timer.watch("build kg object index"):
            kg_object_index = KGObjectIndex.from_entities(
                list(wdentity_ids.intersection(wdentities.keys())),
                wdentities,
                wdprops,
                n_hop=self.cfg.data_graph.max_n_hop,
                traversal_option=TraversalOption.TransitiveOnly,
            )

        with self.timer.watch("build dg & sg"):
            dg_factory = DGFactory(wdentities, wdprops)
            dg = dg_factory.create_dg(
                table, kg_object_index, max_n_hop=self.cfg.data_graph.max_n_hop
            )
            cg_factory = CGFactory(wdentities, wdentity_labels, wdclasses, wdprops)
            cg = cg_factory.create_cg(table, dg)

        with self.timer.watch("run inference"):
            if self.cfg.psl.experiment_model:
                logger.debug("Using experiment PSL model")
                cls = partial(
                    {
                        "exp": PSLGramModelExp,
                        "exp2": PSLGramModelExp2,
                    }[self.cfg.psl.experiment_model],
                    wdprop_domains=self.wdprop_domains,
                    wdprop_ranges=self.wdprop_ranges,
                )
            else:
                cls = PSLGramModel

            edge_probs, cta_probs = cls(
                wdentities=wdentities,
                wdentity_labels=self.wdentity_labels,
                wdclasses=wdclasses,
                wdprops=wdprops,
                wd_numprop_stats=self.wd_numprop_stats,
                disable_rules=self.cfg.psl.disable_rules,
            ).predict(table, cg, dg, verbose=verbose, debug=False)

            edge_probs = PSLModel.normalize_probs(edge_probs, eps=self.cfg.psl.eps)

            if self.cfg.psl.postprocessing == "steiner_tree":
                pp = SteinerTree(table, cg, dg, edge_probs, self.cfg.psl.threshold)
            elif self.cfg.psl.postprocessing == "arborescence":
                pp = MinimumArborescence(
                    table, cg, dg, edge_probs, self.cfg.psl.threshold
                )
            elif self.cfg.psl.postprocessing == "simplepath":
                pp = PostProcessingSimplePath(
                    table, cg, dg, edge_probs, self.cfg.psl.threshold
                )
            elif self.cfg.psl.postprocessing == "pairwise":
                pp = PairwiseSelection(
                    table, cg, dg, edge_probs, self.cfg.psl.threshold
                )
            else:
                raise NotImplementedError(self.cfg.psl.postprocessing)

            pred_cpa = pp.get_result()
            pred_cta = {
                ci: max(classes.items(), key=itemgetter(1))[0]
                for ci, classes in cta_probs.items()
            }

        sm_helper = WikidataSemanticModelHelper(
            wdentities, wdentity_labels, wdclasses, wdprops
        )
        sm = sm_helper.create_sm(table, pred_cpa, pred_cta)
        sm = sm_helper.minify_sm(sm)
        return Annotation(
            sm=sm,
            dg=dg,
            cg=cg,
            cg_edge_probs=edge_probs,
            cta_probs=cta_probs,
            pred_cpa=pred_cpa,
            pred_cta=pred_cta,
        )

    def get_entities(
        self, wdentity_ids: Set[str], n_hop: int = 1, verbose: bool = False
    ) -> Dict[str, WDEntity]:
        assert n_hop <= 2
        batch_size = 30
        wdentities: Dict[str, WDEntity] = {}
        pp = Parallel()
        for wdentity_id in wdentity_ids:
            wdentity = self.wdentities.get(wdentity_id, None)
            if wdentity is not None:
                wdentities[wdentity_id] = wdentity

        if isinstance(self.wdentities, WDProxyDB):
            missing_qnode_ids = [
                wdentity_id
                for wdentity_id in wdentity_ids
                if wdentity_id not in wdentities
                and not self.wdentities.does_not_exist_locally(wdentity_id)
            ]
            if len(missing_qnode_ids) > 0:
                resp = pp.map(
                    query_wikidata_entities,
                    [
                        missing_qnode_ids[i : i + batch_size]
                        for i in range(0, len(missing_qnode_ids), batch_size)
                    ],
                    show_progress=verbose,
                    progress_desc=f"query wikidata for get missing entities in hop 1",
                    is_parallel=True,
                )
                for odict in resp:
                    for k, v in odict.items():
                        wdentities[k] = v
                        self.wdentities[k] = v

        if n_hop > 1:
            next_wdentity_ids = set()
            for wdentity in wdentities.values():
                for p, stmts in wdentity.props.items():
                    for stmt in stmts:
                        if stmt.value.is_qnode(stmt.value):
                            next_wdentity_ids.add(stmt.value.as_entity_id())
                        for qvals in stmt.qualifiers.values():
                            next_wdentity_ids = next_wdentity_ids.union(
                                qval.as_entity_id()
                                for qval in qvals
                                if qval.is_qnode(qval)
                            )
            next_wdentity_ids = list(next_wdentity_ids.difference(wdentities.keys()))
            for wdentity_id in tqdm(
                next_wdentity_ids,
                desc="load entities in 2nd hop from db",
                disable=not verbose,
            ):
                wdentity = self.wdentities.get(wdentity_id, None)
                if wdentity is not None:
                    wdentities[wdentity_id] = wdentity

            if isinstance(self.wdentities, WDProxyDB):
                next_wdentity_ids = [
                    qnode_id
                    for qnode_id in next_wdentity_ids
                    if qnode_id not in wdentities
                    and not self.wdentities.does_not_exist_locally(qnode_id)
                ]
                if len(next_wdentity_ids) > 0:
                    resp = pp.map(
                        query_wikidata_entities,
                        [
                            next_wdentity_ids[i : i + batch_size]
                            for i in range(0, len(next_wdentity_ids), batch_size)
                        ],
                        show_progress=verbose,
                        progress_desc=f"query wikidata for get missing entities in hop {n_hop}",
                        is_parallel=True,
                    )
                    for odict in resp:
                        for k, v in odict.items():
                            wdentities[k] = v
                            self.wdentities[k] = v
        return wdentities

    def get_entity_labels(
        self, wdentities: Dict[str, WDEntity], verbose: bool = False
    ) -> Dict[str, str]:
        id2label: Dict[str, str] = {}
        for qnode in tqdm(wdentities.values(), disable=not verbose, desc=""):
            qnode: WDEntity
            id2label[qnode.id] = str(qnode.label)
            for stmts in qnode.props.values():
                for stmt in stmts:
                    if stmt.value.is_qnode(stmt.value):
                        qnode_id = stmt.value.as_entity_id()
                        if qnode_id in self.wdentity_labels:
                            label = self.wdentity_labels[qnode_id].label
                        else:
                            label = qnode_id
                        id2label[qnode_id] = label
                    for qvals in stmt.qualifiers.values():
                        for qval in qvals:
                            if qval.is_qnode(qval):
                                qnode_id = qval.as_entity_id()
                                if qnode_id in self.wdentity_labels:
                                    label = self.wdentity_labels[qnode_id].label
                                else:
                                    label = qnode_id
                                id2label[qnode_id] = label
        return id2label



from grams.algorithm.postprocessing.simple_path import PostProcessingSimplePath
from grams.algorithm.postprocessing.steiner_tree import SteinerTree
from grams.algorithm.postprocessing.arborescence import MinimumArborescence
from grams.algorithm.postprocessing.known_targets import KnownTargets
from grams.algorithm.postprocessing.pairwise_selection import PairwiseSelection



import copy
from dataclasses import dataclass
from operator import attrgetter, itemgetter

import networkx as nx
import matplotlib.pyplot as plt
from graph.interface import IMultiGraph
import sm.misc as M
from typing import (
    Generic,
    Set,
    List,
    Dict,
    Tuple,
    Callable,
    FrozenSet,
    Optional,
    Any,
    TypeVar,
)
from collections import defaultdict

"""
Find a top-k approximate minimum steiner arborescence using the BANK algorithm
"""


@dataclass
class Edge:
    __slots__ = ("id", "source_id", "target_id", "edge_key", "weight", "n_edges")
    id: str
    source_id: str
    target_id: str
    edge_key: str
    weight: float
    n_edges: int


G = TypeVar("G", bound=IMultiGraph)


@dataclass
class Solution(Generic[G]):
    id: FrozenSet[str]
    graph: G
    weight: float

    def get_n_edges(self):
        if not hasattr(self, "n_edges"):
            self.n_edges = sum(e.n_edges for _, _, e in self.graph.iter_edges())
        return self.n_edges


class NoSingleRootException(Exception):
    pass


# noinspection PyMethodMayBeStatic
class SteinerTreeBankSolver:
    def __init__(
        self,
        original_graph: nx.MultiDiGraph,
        terminal_nodes: Set[str],
        weight_fn: Callable[[dict], float],
        solution_cmp_fn: Optional[Callable[[Solution], Any]] = None,
        top_k_st: int = 10,
        top_k_path: int = 10,
        allow_shorten_graph: bool = True,
    ):
        # original graph
        self.original_graph = original_graph
        # function that extract weights
        self.weight_fn = weight_fn
        # function to compare & sort solutions
        self.solution_cmp_fn = solution_cmp_fn or attrgetter("weight")
        # graph that the bank algorithm will work with
        self.graph: nx.MultiDiGraph = None
        # output graphs
        self.solutions: List[Solution] = []
        self.terminal_nodes = terminal_nodes
        self.top_k_st = top_k_st
        self.top_k_path = top_k_path
        self.allow_shorten_graph = allow_shorten_graph

    def run(self):
        self.graph = self._preprocessing(self.original_graph, self.weight_fn)

        if nx.is_weakly_connected(self.graph):
            self.solutions = self._solve(
                self.graph, self.terminal_nodes, self.top_k_st, self.top_k_path
            )
        else:
            graphs = self._split_connected_components(self.graph)
            final_solutions = None
            for g in graphs:
                terminal_nodes = self.terminal_nodes.intersection(list(g.nodes))
                solutions = self._solve(
                    g, terminal_nodes, self.top_k_st, self.top_k_path
                )
                if final_solutions is None:
                    final_solutions = solutions
                else:
                    next_solutions = []
                    for current_sol in final_solutions:
                        for sol in solutions:
                            next_solutions.append(
                                self._merge_graph(current_sol.graph, sol.graph)
                            )
                    final_solutions = self._sort_solutions(
                        next_solutions, return_solution=True
                    )[: self.top_k_st]

            if final_solutions is not None:
                self.solutions = final_solutions

        # [self._get_roots(sol.graph) for sol in self.solutions]
        # [sol.weight for sol in self.solutions]
        return [
            self._postprocessing(self.original_graph, self.graph, sol.graph)
            for sol in self.solutions
        ], self.solutions

    def _preprocessing(self, g: nx.MultiDiGraph, weight_fn: Callable[[dict], float]):
        ng = nx.MultiDiGraph()

        # convert edges
        edge_id_count = 0
        for uid, vid, eid, edata in g.edges(data=True, keys=True):
            ng.add_edge(
                uid,
                vid,
                key=eid,
                data=Edge(
                    id=str(edge_id_count),
                    source_id=uid,
                    target_id=vid,
                    edge_key=eid,
                    weight=weight_fn(edata),
                    n_edges=1,
                ),
            )
            edge_id_count += 1

        if self.allow_shorten_graph:
            # shorten path using the following heuristic
            # for a node that only connect two nodes (indegree & outdegree = 1) and not terminal nodes, we replace the node by one edge
            # map from the replaced edge to the removed nodes
            removed_nodes = {}
            for uid in list(ng.nodes):
                if (
                    uid not in self.terminal_nodes
                    and g.in_degree(uid) == 1
                    and g.out_degree(uid) == 1
                ):
                    inedge: Edge = next(iter(ng.in_edges(uid, data="data")))[-1]
                    outedge: Edge = next(iter(ng.out_edges(uid, data="data")))[-1]
                    if inedge.edge_key == outedge.edge_key:
                        new_edge_key = inedge.edge_key
                    else:
                        new_edge_key = f"{inedge.edge_key}-{outedge.edge_key}"
                    if not ng.has_edge(
                        inedge.source_id, outedge.target_id, key=new_edge_key
                    ):
                        # replace it only if we don't have that link before
                        ng.remove_node(uid)
                        ng.add_edge(
                            inedge.source_id,
                            outedge.target_id,
                            key=new_edge_key,
                            data=Edge(
                                id=str(edge_id_count),
                                source_id=inedge.source_id,
                                target_id=outedge.target_id,
                                edge_key=new_edge_key,
                                weight=inedge.weight + outedge.weight,
                                n_edges=2,
                            ),
                        )
                        removed_nodes[str(edge_id_count)] = (uid, inedge, outedge)
                        edge_id_count += 1

            # store the removed nodes to restore it later
            ng.graph["removed_nodes"] = removed_nodes
        return ng

    def _postprocessing(
        self,
        origin_graph: nx.MultiDiGraph,
        prep_graph: nx.MultiDiGraph,
        out_graph: nx.MultiDiGraph,
    ):
        if self.allow_shorten_graph:
            removed_nodes = prep_graph.graph["removed_nodes"]
        else:
            removed_nodes = set()

        g = origin_graph.copy()
        selected_edges = set()
        for uid, vid, edge in out_graph.edges(data="data"):
            if edge.id in removed_nodes:
                removed_node, inedge, outedge = removed_nodes[edge.id]
                selected_edges.add(
                    (inedge.source_id, inedge.target_id, inedge.edge_key)
                )
                selected_edges.add(
                    (outedge.source_id, outedge.target_id, outedge.edge_key)
                )
            else:
                selected_edges.add((uid, vid, edge.edge_key))

        for uid, vid, eid in list(g.edges(keys=True)):
            if (uid, vid, eid) not in selected_edges:
                g.remove_edge(uid, vid, eid)
        for uid in list(g.nodes):
            if g.in_degree(uid) == 0 and g.out_degree(uid) == 0:
                g.remove_node(uid)
        return g

    def _merge_graph(self, g1: nx.MultiDiGraph, g2: nx.MultiDiGraph) -> nx.MultiDiGraph:
        g = g1.copy()
        for uid, vid, eid, edata in g2.edges(keys=True, data=True):
            g.add_edge(uid, vid, key=eid, **edata)
        return g

    def _split_connected_components(self, g: nx.MultiDiGraph):
        connected_comps = [
            comp
            for comp in nx.weakly_connected_components(g)
            # must have at least two terminal nodes (to form a graph)
            if len(self.terminal_nodes.intersection(comp)) > 1
        ]

        node2comp = {}
        for i, comp in enumerate(connected_comps):
            for uid in comp:
                node2comp[uid] = i

        subgs = [nx.MultiDiGraph() for _ in range(len(connected_comps))]
        for uid, vid, eid, edata in g.edges(keys=True, data=True):
            if uid not in node2comp:
                continue
            subgs[node2comp[uid]].add_edge(uid, vid, key=eid, **edata)
        return subgs

    def _solve(
        self,
        g: nx.MultiDiGraph,
        terminal_nodes: Set[str],
        top_k_st: int,
        top_k_path: int,
    ):
        """Despite the name, this is finding steiner tree. Assuming their is a root node that connects all
        terminal nodes together.
        """
        roots = set(g.nodes.keys())

        attr_visit_hists = []
        # to ensure the order
        for uid in list(sorted(terminal_nodes)):
            visit_hist = UpwardTraversal.top_k_beamsearch(g, uid, top_k_path)
            roots = roots.intersection(visit_hist.paths.keys())
            attr_visit_hists.append((uid, visit_hist))

        if len(roots) == 0:
            # there is no nodes that can connect to all terminal nodes either this are disconnected
            # components or you pass a directed graph with weakly connected components (a -> b <- c)
            if nx.is_weakly_connected(g):
                # perhaps, we can break the weakly connected components by breaking one of the link (a -> b <- c)
                raise NoSingleRootException(
                    "You pass a weakly connected component and there are parts of the graph like this (a -> b <- c). Fix it before running this algorithm"
                )
            raise Exception(
                "Your graph is disconnected. Consider splitting them before calling bank solver"
            )

        # to ensure the order again & remove randomness
        roots = sorted(roots)

        # merge the paths using beam search
        results = []
        for root in roots:
            current_states = []
            uid, visit_hist = attr_visit_hists[0]
            for path in visit_hist.paths[root]:
                pg = nx.MultiDiGraph()
                if len(path.path) > 0:
                    assert uid == path.path[0].target_id
                pg.add_node(uid)
                for e in path.path:
                    pg.add_node(e.source_id)
                    pg.add_edge(
                        e.source_id, e.target_id, e.edge_key, weight=e.weight, data=e
                    )
                current_states.append(pg)

            if len(current_states) > top_k_st:
                current_states = self._sort_solutions(current_states)[:top_k_st]

            for uid, visit_hist in attr_visit_hists[1:]:
                next_states = []
                for state in current_states:
                    for path in visit_hist.paths[root]:
                        pg = state.copy()
                        if len(path.path) > 0:
                            assert uid == path.path[0].target_id
                        pg.add_node(uid)
                        for e in path.path:
                            if e.source_id not in pg.nodes:
                                pg.add_node(e.source_id)
                            # TODO: here we don't check by edge_key because we may create another edge of different key
                            # hope this new path has been exploited before.
                            if not pg.has_edge(e.source_id, e.target_id):
                                pg.add_edge(
                                    e.source_id,
                                    e.target_id,
                                    key=e.edge_key,
                                    weight=e.weight,
                                    data=e,
                                )

                        # if there are more than path between two nodes within
                        # two hop, we'll select one
                        update_graph = False
                        for nid in pg.nodes:
                            if pg.in_degree(nid) >= 2:
                                grand_parents = defaultdict(list)
                                for inedge in pg.in_edges(nid, keys=True, data="data"):
                                    grand_parents[inedge[0]].append((inedge,))
                                    for grand_inedge in pg.in_edges(
                                        inedge[0], keys=True, data="data"
                                    ):
                                        grand_parents[grand_inedge[0]].append(
                                            (grand_inedge, inedge)
                                        )

                                for grand_parent, edges in grand_parents.items():
                                    if len(edges) > 1:
                                        # we need to select one path from this grand parent to the rest
                                        # they have the same length, so we select the one has smaller weight
                                        edges = sorted(
                                            edges,
                                            key=lambda x: x[0][-1].weight
                                            + x[1][-1].weight
                                            if len(x) == 2
                                            else x[0][-1].weight * 2,
                                        )

                                        for lst in edges[1:]:
                                            for edge in lst:
                                                # TODO: handle removing edges multiple times
                                                try:
                                                    pg.remove_edge(
                                                        edge[0], edge[1], edge[2]
                                                    )
                                                except:
                                                    continue
                                        update_graph = True
                        if update_graph:
                            for nid in list(pg.nodes):
                                if pg.in_degree(nid) == 0 and pg.out_degree(nid) == 0:
                                    pg.remove_node(nid)
                        # after add a path to the graph, it can create new cycle, detect and fix it
                        try:
                            # note that the function I'm using doesn't return parallel edges
                            cycles_iter = nx.find_cycle(pg, uid, orientation="reverse")
                            # we can show that if the graph contain cycle, there is a better path
                            #
                            # cycles_iter = [(uid, vid) for uid, vid, eid, orien in cycles_iter]
                            # for _g in self._break_cycles(root, pg, cycles_iter):
                            #     next_states.append(_g)
                        except nx.NetworkXNoCycle:
                            next_states.append(pg)
                        if len(list(pg.edges)) != len(
                            {(uid, vid) for uid, vid, eid in pg.edges(keys=True)}
                        ):
                            assert False
                if len(next_states) > top_k_st:
                    next_states = self._sort_solutions(next_states)[:top_k_st]
                current_states = next_states
                # cgs = [g for g in next_states if len(list(nx.simple_cycles(g))) > 0]
                # nx.draw_networkx(cgs[0]); plt.show()
                # nx.draw(cgs[0]); plt.show()
            results += current_states

        return self._sort_solutions(results, return_solution=True)

    def _break_cycles(
        self, root: str, g: nx.MultiDiGraph, cycles_iter: List[Tuple[str, str]]
    ):
        # g = current_states[0]; g = self.output_graphs[0]
        # pos = nx.kamada_kawai_layout(g); nx.draw_networkx(g, pos); nx.draw_networkx_edge_labels(g, pos, edge_labels={(u, v): d for u, v, d in g.edges(keys=True)}); plt.show()
        # nx.draw(g); plt.show()
        return self._break_cycles_brute_force(root, g, cycles_iter)

    def _break_cycles_brute_force(
        self, root: str, g: nx.MultiDiGraph, cycles_iter: List[Tuple[str, str]]
    ):
        # one side effect of this approach is that it may separate the graph
        # currently we say it only happen when the target node of the remove edge only has one incoming edge
        # if it has two edges, then the other source (not of the remove edge) must have a path from root -> itself
        # now, since we have cascade removing, if the path doesn't go through the removed node, then it's okay,
        # if the path goes through the remove node, it's impossible since we only remove node that indegree == 0 or
        # outdegree == 0
        parallel_edges = []
        for uid, vid in cycles_iter:
            if g.in_degree(vid) == 1:
                # we can't remove this edge, as removing it will make it's unreachable from the root
                # so just skip edge
                continue
            edges = [edata["data"] for eid, edata in g[uid][vid].items()]
            edge_weight = min(edges, key=attrgetter("weight")).weight
            parallel_edges.append((uid, vid, edges, edge_weight))

        # not comparing based on weight anymore since psl can be quite difficult to select correct edge
        # min_edge_weight = min(parallel_edges, key=itemgetter(3))[3]
        # unbreakable = {i for i, x in enumerate(parallel_edges) if x[3] == min_edge_weight}
        # new_graphs = []
        # if len(unbreakable) < len(parallel_edges):
        #     # it's great! we have some edge to break!
        #     # for each breakable edge, we will have a new graph
        #     for i, item in enumerate(parallel_edges):
        #         if i in unbreakable:
        #             continue
        #         ng: nx.MultiDiGraph = g.copy()
        #         ng.remove_edge(item[0], item[1])
        #         ng = self._remove_redundant_nodes(root, ng)
        #         new_graphs.append(ng)
        # else:
        # so bad, we have to try one by one
        new_graphs = []
        for uid, vid, edges, edge_weight in parallel_edges:
            ng: nx.MultiDiGraph = g.copy()
            ng.remove_edge(uid, vid)
            # self._draw(ng); self._draw(g)
            ng = self._remove_redundant_nodes(root, ng)
            new_graphs.append(ng)

        # just assert if it works as expected
        for ng in new_graphs:
            assert len(list(nx.simple_cycles(ng))) == 0
        return new_graphs

    def _sort_solutions(
        self, graphs: List[nx.MultiDiGraph], return_solution: bool = False
    ):
        """Sort the solution by th"""
        solutions = {}
        for g in graphs:
            id = frozenset((e.id for _, _, e in g.edges(data="data")))
            if id in solutions:
                continue

            weight = sum(e.weight for _, _, e in g.edges(data="data"))
            solutions[id] = Solution(id, g, weight)

        solutions = sorted(solutions.values(), key=self.solution_cmp_fn)
        if return_solution:
            return solutions
        return [sol.graph for sol in solutions]

    def _remove_redundant_nodes(self, root: str, g: nx.MultiDiGraph):
        # remove nodes in the graph that shouldn't be in a steiner tree
        while True:
            removed_nodes = []
            for uid in g.nodes:
                if uid == root or uid in self.terminal_nodes:
                    continue
                if g.in_degree(uid) == 0 or g.out_degree(uid) == 0:
                    removed_nodes.append(uid)
            if len(removed_nodes) == 0:
                break
            for uid in removed_nodes:
                g.remove_node(uid)
        return g

    def _get_roots(self, g: nx.MultiDiGraph):
        """This function is mainly used for debugging"""
        return [uid for uid in g if g.in_degree(uid) == 0]

    @staticmethod
    def _draw(g: nx.MultiDiGraph):
        """This function is mainly used for debugging"""
        pos = nx.kamada_kawai_layout(g)
        nx.draw_networkx(g, pos)
        nx.draw_networkx_edge_labels(
            g, pos, edge_labels={(u, v): d for u, v, d in g.edges(keys=True)}
        )
        plt.show()


class CycleBreaker:
    def spanning_arborescence(
        self, g: nx.MultiDiGraph, root: str, terminal_nodes: Set[str]
    ):
        resp = nx.algorithms.tree.branchings.minimum_spanning_arborescence(
            g, attr="weight", default=None, preserve_attrs=True
        )
        ng = nx.MultiDiGraph()
        for uid, vid, edge in resp.edges(data="data"):
            ng.add_edge(uid, vid, key=edge.edge_key, weight=edge.weight, data=edge)
        # after breaking we may have multiple root nodes or leaf so we just need to clean them
        return [self._truncate_redundant_nodes(ng, root, terminal_nodes)]

    def one_edge_at_a_time(
        self, og: nx.MultiDiGraph, root: str, terminal_nodes: Set[str]
    ):
        """This approach doesn't allow us to select the graph that has shorter length due to multiple cycles in the graph"""
        current_graphs = [og]
        has_cycle = True
        while has_cycle:
            # loop until we don't have any cycle left
            new_graphs = []
            for g in current_graphs:
                for cycle in nx.simple_cycles(g):
                    cycle.append(cycle[0])
                    edge_groups = []
                    for i in range(len(cycle) - 1):
                        uid = cycle[i]
                        vid = cycle[i + 1]
                        # we may have more than one edge between two nodes
                        edges = [edata["data"] for eid, edata in g[uid][vid].items()]
                        edge_weight = min(edges, key=attrgetter("weight"))
                        edge_groups.append((uid, vid, edges, edge_weight))

                    min_edge_weight = min(edge_groups, itemgetter(3))
                    unbreakable = {
                        i for i, x in enumerate(edge_groups) if x[3] == min_edge_weight
                    }
                    if len(unbreakable) < len(edge_groups):
                        # it's great! we have some edge to break!
                        # for each breakable edge, we will have a new graph
                        for i, item in enumerate(edge_groups):
                            if i in unbreakable:
                                continue
                            ng: nx.MultiDiGraph = g.copy()
                            # we remove all instead of just one
                            ng.remove_edge(item[0], item[1])
                            ng = self._truncate_redundant_nodes(
                                ng, root, terminal_nodes
                            )
                            new_graphs.append(ng)
                    else:
                        # so bad, we have to try one by one
                        tmp = []
                        for uid, vid, edges, edge_weight in edge_groups:
                            ng: nx.MultiDiGraph = g.copy()
                            ng.remove_edge(uid, vid)
                            ng = self._truncate_redundant_nodes(
                                ng, root, terminal_nodes
                            )
                            # given a score for each split, we prefer the one that is smaller?
                            # however, we can't calculate the depth since the graph may contain multiple cycles
                            # this make this approach not good.
                            tmp.append(ng)

                        # find which one is better

                    pass

    def _truncate_redundant_nodes(
        self, g: nx.MultiDiGraph, root: str, terminal_nodes: Set[str]
    ):
        while True:
            removed_nodes = []
            for uid in g.nodes:
                if uid == root or uid in terminal_nodes:
                    continue
                if g.in_degree(uid) == 0 or g.out_degree(uid) == 0:
                    removed_nodes.append(uid)
            if len(removed_nodes) == 0:
                break
            for uid in removed_nodes:
                g.remove_node(uid)
        return g


@dataclass
class UpwardPath:
    __slots__ = ("visited_nodes", "path", "weight")

    visited_nodes: Set[str]
    path: List[Edge]
    weight: float

    @staticmethod
    def empty(start_node_id: str):
        return UpwardPath({start_node_id}, [], 0.0)

    def push(self, edge: Edge) -> "UpwardPath":
        c = self.clone()
        c.path.append(edge)
        c.visited_nodes.add(edge.source_id)
        c.weight += edge.weight
        return c

    def clone(self):
        return UpwardPath(
            copy.copy(self.visited_nodes), copy.copy(self.path), self.weight
        )


@dataclass
class UpwardTraversal:
    __slots__ = ("source_id", "paths")

    source_id: str

    # storing that we can reach this node through those list of paths
    paths: Dict[str, List[UpwardPath]]

    @staticmethod
    def top_k_beamsearch(g: nx.MultiDiGraph, start_node_id: str, top_k_path: int):
        travel_hist = UpwardTraversal(start_node_id, dict())
        travel_hist.paths[start_node_id] = [UpwardPath.empty(start_node_id)]
        for source_id, target_id, edge_id, orientation in nx.edge_bfs(
            g, start_node_id, orientation="reverse"
        ):
            if source_id not in travel_hist.paths:
                travel_hist.paths[source_id] = []

            edge: Edge = g.edges[source_id, target_id, edge_id]["data"]
            #
            # for edata in g[source_id][target_id].values():
            #     edge: Edge = edata['data']
            for path in travel_hist.paths[target_id]:
                if source_id in path.visited_nodes:
                    # path will become loopy, which we don't want to do
                    continue
                path = path.push(edge)
                travel_hist.paths[source_id].append(path)

            # we trim the number of paths in here
            if len(travel_hist.paths[source_id]) > top_k_path:
                # calculate the score of each path, and then select the best one
                travel_hist.sort_paths(source_id)
                travel_hist.paths[source_id] = travel_hist.paths[source_id][:top_k_path]
        return travel_hist

    def sort_paths(self, node_id: str):
        self.paths[node_id] = sorted(self.paths[node_id], key=attrgetter("weight"))



from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Mapping, Optional, Tuple, Set
import networkx as nx
from kgdata.wikidata.models import WDEntity, WDValue, WDProperty, WDClass
from grams.algorithm.data_graph.dg_graph import (
    CellNode,
    DGEdge,
    DGGraph,
    DGNode,
    DGStatementID,
    EdgeFlowSource,
    EdgeFlowTarget,
    StatementNode,
    DGPath,
    DGPathEdge,
    EntityValueNode,
    DGPathNodeStatement,
    DGPathNodeEntity,
    DGPathNodeLiteralValue,
    DGPathExistingNode,
    FlowProvenance,
)


class KGInference:
    def __init__(
        self,
        dg: DGGraph,
        wdentities: Mapping[str, WDEntity],
        wdprops: Mapping[str, WDProperty],
    ):
        # mapping from qnode id, and property to a list of statement values (with the corresponding DG statement node if exist)
        # the reason we want to include all is that we want to know if we need to add new statement value or reuse existing value
        self.subkg: Dict[
            Tuple[str, str], List[Tuple[Optional[StatementNode], WDValue]]
        ] = {}
        self.wdentities = wdentities
        self.wdprops = wdprops
        self.dg = dg

        for s in dg.iter_nodes():
            if not isinstance(s, StatementNode):
                continue

            dgsid = DGStatementID.parse_id(s.id)
            self._set_stmt_node(s.qnode_id, dgsid.predicate, dgsid.statement_index, s)

    def infer_subproperty(self):
        """Infer new properties via sub-property of (inheritance)"""
        properties = set()
        qualifiers = set()

        for u in self.dg.iter_nodes():
            if isinstance(u, StatementNode):
                continue

            for s in self.dg.successors(u.id):
                assert isinstance(s, StatementNode)
                properties.add(s.predicate)

                for v in self.dg.successors(s.id):
                    sv_edges = self.dg.get_edges_between_nodes(s.id, v.id)
                    qualifiers.update(
                        (
                            sv_edge.predicate
                            for sv_edge in sv_edges
                            if sv_edge != s.predicate
                        )
                    )

        parent_props = self._build_parent_map(properties)
        parent_qualifiers = self._build_parent_map(qualifiers)

        # list of new properties and qualifiers that we will infer
        new_props = []
        new_qualifiers = []

        for s in self.dg.iter_nodes():
            if not isinstance(s, StatementNode):
                continue

            # parents of the statement
            # parents = {u.id for u in self.dg.predecessors(s.id)}
            prop = s.predicate

            # children that are properties of the statement
            prop_children = set()
            # children that are qualifiers of the statement
            qualifier2children = defaultdict(set)
            for e in self.dg.out_edges(s.id):
                if e.predicate == prop:
                    prop_children.add(e.target)
                else:
                    qualifier2children[e.predicate].add(e.target)

            stmt_index = DGStatementID.parse_id(s.id).statement_index
            if prop in parent_props:
                for parent_prop in parent_props[prop]:
                    for vid in prop_children:
                        for source_flow, flow_provenances in s.iter_source_flow(
                            EdgeFlowTarget(vid, prop)
                        ):
                            new_prop = InferredNewProp(
                                qnode_id=s.qnode_id,
                                new_prop=parent_prop,
                                value=self._get_stmt_value(
                                    s.qnode_id, prop, stmt_index
                                ),
                                source_id=source_flow.source_id,
                                target_id=vid,
                                qualifier_edges=[],
                                flow_provenances=flow_provenances,
                            )
                            new_props.append(new_prop)

            for q, children in qualifier2children.items():
                for pq in parent_qualifiers.get(q, []):
                    for vid in children:
                        for source_flow, provenance in s.iter_source_flow(
                            EdgeFlowTarget(vid, q)
                        ):
                            new_qualifiers.append(
                                InferredNewQualifier(
                                    statement_id=s.id,
                                    new_qualifier=pq,
                                    source_id=source_flow.source_id,
                                    property=source_flow.edge_id,
                                    target_id=vid,
                                    flow_provenances=provenance,
                                )
                            )

        self.add_inference(new_props, new_qualifiers)
        return self

    def kg_transitive_inference(self):
        """Infer new relationship based on the transitive property: a -> b -> c => a -> c"""

        # find the list of transitive properties in the graph
        transitive_props = set()
        for e in self.dg.edges():
            prop = self.wdprops[e.predicate]
            # transitive class
            if "Q18647515" in prop.instanceof:
                transitive_props.add(e.predicate)

        # now start from node u, we find if there is another v connect to u via a transitive property, and another p connect
        # to v with the same property, we don't need to keep the chain going as even if it's longer, we will eventually loop
        # through all item in the chain by looping through nodes in the graph

        chains = []
        for u in self.dg.iter_nodes():
            if isinstance(u, StatementNode):
                continue

            for stmt, us_edges in self.dg.group_out_edges(u.id):
                assert isinstance(stmt, StatementNode)
                for trans_prop in transitive_props:
                    if trans_prop not in us_edges:
                        continue
                    us_edge = us_edges[trans_prop]
                    for v, sv_edges in self.dg.group_out_edges(stmt.id):
                        if trans_prop not in sv_edges:
                            continue
                        sv_edge = sv_edges[trans_prop]
                        if not stmt.is_same_flow(us_edge, sv_edge):
                            # don't allow infer new link across rows
                            continue

                        for stmt2, vs2_edges in self.dg.group_out_edges(v.id):
                            assert isinstance(stmt2, StatementNode)
                            if trans_prop not in vs2_edges:
                                continue
                            vs2_edge: DGEdge = vs2_edges[trans_prop]
                            for v2, s2v2_edges in self.dg.group_out_edges(stmt2.id):
                                if trans_prop not in s2v2_edges:
                                    continue
                                s2v2_edge = s2v2_edges[trans_prop]
                                if not stmt2.is_same_flow(vs2_edge, s2v2_edge):
                                    # don't allow infer new link across rows
                                    continue

                                # we now record the chain
                                chains.append((us_edge, sv_edge, vs2_edge, s2v2_edge))

        # make sure that there is no duplication in the chains
        assert len(chains) == len(
            {
                (
                    us_edge.source,
                    us_edge.predicate,
                    us_edge.target,
                    sv_edge.source,
                    sv_edge.predicate,
                    sv_edge.target,
                    vs2_edge.source,
                    vs2_edge.predicate,
                    vs2_edge.target,
                    s2v2_edge.source,
                    s2v2_edge.predicate,
                    s2v2_edge.target,
                )
                for us_edge, sv_edge, vs2_edge, s2v2_edge in chains
            }
        )

        # generate new property, but qualifiers cannot inherit via transitive inference
        new_props = []
        for us_edge, sv_edge, vs2_edge, s2v2_edge in chains:
            trans_prop = us_edge.predicate
            stmt = self.dg.get_node(us_edge.target)
            stmt2 = self.dg.get_node(vs2_edge.target)
            assert isinstance(stmt, StatementNode) and isinstance(stmt2, StatementNode)

            prop_value = self._get_stmt_value(
                stmt2.qnode_id,
                trans_prop,
                DGStatementID.parse_id(stmt2.id).statement_index,
            )
            # prop_value = stmt2.qnode.props[trans_prop][DGStatementID.parse_id(stmt2.id).statement_index].value

            # calculating the provenance of the new transitive link. however, I don't know in case we have multiple
            # provenance since it depends on both first leg and second leg. for now, we put an assertion to handle
            # only the case where we have two legs has the same provenance, which make the new transitive link has
            # the same provenance too.
            first_leg_provenances = stmt.get_provenance_by_edge(us_edge, sv_edge)
            second_leg_provenances = stmt2.get_provenance_by_edge(vs2_edge, s2v2_edge)
            assert (
                len(first_leg_provenances) == 1
                and len(second_leg_provenances) == 1
                and first_leg_provenances[0] == second_leg_provenances[0]
            )

            provenances = first_leg_provenances
            new_props.append(
                InferredNewProp(
                    qnode_id=stmt.qnode_id,
                    new_prop=trans_prop,
                    value=prop_value,
                    source_id=us_edge.source,
                    target_id=s2v2_edge.target,
                    qualifier_edges=[],
                    flow_provenances=provenances,
                )
            )

        self.add_inference(new_props, [])
        return self

    def add_inference(
        self,
        new_props: List[InferredNewProp],
        new_qualifiers: List[InferredNewQualifier],
    ):
        """After we run inference, we got a list of new properties and new qualifiers that can update using this function"""
        new_nodes = []
        new_edges = []

        # here we enforce the constraint that there is no cross links
        # between nodes in different rows, this happen because transitive
        # inference generate links cross rows
        new_props = [
            new_prop
            for new_prop in new_props
            if DGEdge.can_link(
                self.dg.get_node(new_prop.source_id),
                self.dg.get_node(new_prop.target_id),
            )
        ]
        new_qualifiers = [
            new_qualifier
            for new_qualifier in new_qualifiers
            if DGEdge.can_link(
                self.dg.get_node(new_qualifier.source_id),
                self.dg.get_node(new_qualifier.target_id),
            )
        ]

        for new_prop in new_props:
            stmt_exist = None
            prop = new_prop.new_prop

            # search for existing statement in the KG
            self._track_property(new_prop.qnode_id, prop)
            for sprime, value in self.subkg[new_prop.qnode_id, prop]:
                if sprime is not None and new_prop.value == value:
                    stmt_exist = sprime
                    break

            # if the statement exist, re-use it
            if stmt_exist is not None:
                sprime = stmt_exist
            else:
                for stmt_index, (sprime, value) in enumerate(
                    self.subkg[new_prop.qnode_id, prop]
                ):
                    if new_prop.value == value:
                        stmt_id = DGStatementID(
                            new_prop.qnode_id, prop, stmt_index
                        ).get_id()
                        sprime = StatementNode(
                            stmt_id, new_prop.qnode_id, prop, is_in_kg=True
                        )
                        self._set_stmt_node(new_prop.qnode_id, prop, stmt_index, sprime)
                        new_nodes.append(sprime)
                        break
                else:
                    stmt_index = self.get_next_available_stmt_index(
                        new_prop.qnode_id, prop
                    )
                    stmt_id = DGStatementID(
                        new_prop.qnode_id, prop, stmt_index
                    ).get_id()
                    sprime = StatementNode(
                        stmt_id, new_prop.qnode_id, prop, is_in_kg=False
                    )
                    self._add_stmt_value(
                        new_prop.qnode_id, prop, stmt_index, sprime, new_prop.value
                    )
                new_nodes.append(sprime)

            if not self.dg.has_edge_between_nodes(
                new_prop.source_id, sprime.id, key=prop
            ):
                new_edges.append(
                    DGEdge(
                        source=new_prop.source_id,
                        target=sprime.id,
                        predicate=prop,
                        is_qualifier=False,
                        is_inferred=True,
                    )
                )
            if not self.dg.has_edge_between_nodes(
                sprime.id, new_prop.target_id, key=prop
            ):
                new_edges.append(
                    DGEdge(
                        source=sprime.id,
                        target=new_prop.target_id,
                        predicate=prop,
                        is_qualifier=False,
                        is_inferred=True,
                    )
                )

            sprime.track_provenance(
                EdgeFlowSource(new_prop.source_id, prop),
                EdgeFlowTarget(new_prop.target_id, prop),
                new_prop.flow_provenances,
            )

            # TODO: we haven't add to add add qualifiers, so we assert we don't have any. fix me!
            assert len(new_prop.qualifier_edges) == 0
            # for qual_edge in new_prop.qualifier_edges:
            #     new_edges.append(DGEdge(source=stmt_id, target=qual_edge.target, predicate=qual_edge.predicate, is_qualifier=True, paths=[], is_inferred=True))

        for new_qual in new_qualifiers:
            stmt = self.dg.get_node(new_qual.statement_id)
            assert isinstance(stmt, StatementNode)
            new_edges.append(
                DGEdge(
                    source=new_qual.statement_id,
                    target=new_qual.target_id,
                    predicate=new_qual.new_qualifier,
                    is_qualifier=True,
                    is_inferred=True,
                )
            )
            stmt.track_provenance(
                EdgeFlowSource(new_qual.source_id, new_qual.property),
                EdgeFlowTarget(new_qual.target_id, new_qual.new_qualifier),
                new_qual.flow_provenances,
            )

        for node in new_nodes:
            self.dg.add_node(node)
        for edge in new_edges:
            self.dg.add_edge(edge)

    def _track_property(self, qnode_id: str, prop: str):
        """Ensure that the subkg has values of the qnode's property"""
        if (qnode_id, prop) not in self.subkg:
            lst = []
            for stmt_i, stmt in enumerate(self.wdentities[qnode_id].props.get(prop, [])):
                lst.append((None, stmt.value))
            self.subkg[qnode_id, prop] = lst

    def _get_stmt_value(self, qnode_id: str, prop: str, stmt_index: int):
        self._track_property(qnode_id, prop)
        return self.subkg[qnode_id, prop][stmt_index][1]

    def _set_stmt_node(
        self, qnode_id: str, prop: str, stmt_index: int, stmt: StatementNode
    ):
        self._track_property(qnode_id, prop)
        assert (
            self.subkg[qnode_id, prop][stmt_index][0] is None
        ), "Cannot override existing value in the KG"
        self.subkg[qnode_id, prop][stmt_index] = (
            stmt,
            self.subkg[qnode_id, prop][stmt_index][1],
        )

    def _add_stmt_value(
        self,
        qnode_id: str,
        prop: str,
        stmt_index: int,
        stmt: StatementNode,
        value: WDValue,
    ):
        self._track_property(qnode_id, prop)
        assert stmt_index == len(self.subkg[qnode_id, prop]), "Can only add new value"
        self.subkg[qnode_id, prop].append((stmt, value))

    def get_next_available_stmt_index(self, qnode_id: str, prop: str):
        return len(self.subkg[qnode_id, prop])

    def _build_parent_map(self, props: Set[str]):
        """Build a map from a property to its parents in the same list"""
        parent_props: Dict[str, List[str]] = {}
        for p1 in props:
            parent_props[p1] = []
            for p2 in props:
                if p1 == p2:
                    continue
                if p2 in self.wdprops[p1].ancestors:
                    parent_props[p1].append(p2)
            if len(parent_props[p1]) == 0:
                parent_props.pop(p1)
        return parent_props


@dataclass
class InferredNewProp:
    # information to identify the statement for the new prop (whether to reuse it, or create a new one)
    # the qnode of the new statement (the one will have this property)
    qnode_id: str
    # the new property we inferred
    new_prop: str
    # the value associated with the property (use it to compare if the statement exists) - the reason we use
    # the value instead of the whole statement is that sometimes, desire them to match the qualifier does not
    # sense, if the value exist, then some how the algorithm is already match it so it should be okay
    value: WDValue

    # the source nodes in the data graph
    source_id: str
    # the target node in the data graph that will contain the value of the prop, not the statement node
    target_id: str

    # edges that contain the qualifiers that we want to be copied to the new statement as well
    # the flow of the qualifier can be retrieve
    qualifier_edges: List[DGEdge]

    # flow provenance of the property.
    flow_provenances: List[FlowProvenance]


@dataclass
class InferredNewQualifier:
    # the statement node in the data graph that we will add the qualifier to
    statement_id: str
    # the qualifier we are going to add
    new_qualifier: str

    # source id (node that has the statement id)
    source_id: str
    # the property of the statement (the one that connect source and statement)
    property: str
    # the target node of the qualifier
    target_id: str

    # flow provenance of the qualifier
    flow_provenances: List[FlowProvenance]



import copy
import math
import os
import time
import uuid
from functools import cmp_to_key
from itertools import chain
from multiprocessing import Pool
from operator import attrgetter, itemgetter, xor
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    TypedDict,
    Union,
    cast,
)

import networkx as nx
import sm.misc as M
from grams.algorithm.candidate_graph.cg_graph import (
    CGColumnNode,
    CGEdge,
    CGEntityValueNode,
    CGGraph,
    CGLiteralValueNode,
    CGStatementNode,
)
from grams.algorithm.data_graph.dg_graph import DGGraph, DGNode
from grams.algorithm.deprecated.link_feature import LinkFeatureExtraction

# from grams.algorithm.postprocessing.semtab2020 import SemTab2020PostProcessing
from grams.algorithm.deprecated.type_feature import TypeFeatureExtraction
from grams.inputs.linked_table import LinkedTable

# from grams.algorithm.postprocessing.simple_path import (
#     keep_one_simple_path_between_important_nodes,
# )
from graph.retworkx.api import dag_longest_path
from hugedict.parallel.parallel import Parallel
from kgdata.wikidata.models import (
    WDEntity,
    WDClass,
    WDProperty,
    WDQuantityPropertyStats,
)
from loguru import logger
from networkx.exception import NetworkXUnfeasible
from pslpython.model import Model
from pslpython.partition import Partition
from pslpython.predicate import Predicate
from pslpython.rule import Rule
from steiner_tree.bank import BankSolver, NoSingleRootException, Solution
from tqdm.auto import tqdm

global_objects = {}


class IDMap:
    def __init__(self, counter: int = 0):
        self.counter = counter
        self.map = {}
        self.invert_map = {}

    def add_keys(self, keys: Iterable[Any]):
        for key in keys:
            new_key = f"i-{self.counter}"
            self.map[key] = new_key
            self.invert_map[new_key] = key
            self.counter += 1
        return self

    def m(self, key):
        """Get a new key from old key"""
        return self.map[key]

    def im(self, new_key):
        """Get the old key from the new key"""
        return self.invert_map[new_key]


class FakeIDMap(IDMap):
    def m(self, key):
        return key

    def im(self, new_key):
        return new_key


class PSLConfigs:
    POSTPROCESSING_METHOD = "steiner_tree"
    POSTPROCESSING_STEINER_TREE_FORCE_ADDING_CONTEXT: bool = True


class PSLInference:
    LinkNegPrior = "NegPrior"
    LinkNegParentPropPrior = "NegParentPropPrior"
    CascadingError = "CascadingError"
    FreqLinkOverRow = LinkFeatureExtraction.FreqOverRow
    FreqLinkOverEntRow = LinkFeatureExtraction.FreqOverEntRow
    FreqLinkOverPosLink = LinkFeatureExtraction.FreqOverPossibleLink
    FreqLinkUnmatchOverEntRow = LinkFeatureExtraction.FreqUnmatchOverEntRow
    FreqLinkUnmatchOverPossibleLink = LinkFeatureExtraction.FreqUnmatchOverPossibleLink
    LinkDataTypeMismatch = LinkFeatureExtraction.DataTypeMismatch
    LinkHeaderSimilarity = LinkFeatureExtraction.HeaderSimilarity
    LinkNotFuncDep = LinkFeatureExtraction.NotFuncDep

    TypeNegPrior = "TypeNegPrior"
    TypeNegParentPrior = "TypeNegParentPrior"
    TypeHeaderSimilarity = TypeFeatureExtraction.HeaderSimilarity
    FreqTypeOverRow = TypeFeatureExtraction.FreqOverRow
    FreqTypeInheritOverRow = TypeFeatureExtraction.FreqInheritOverRow
    TypeMustInPropRange = "TypeMustInPropRange"
    TypeOnlyOneConstraint = "TypeOnlyOneConstraint"

    def __init__(
        self,
        qnodes: Mapping[str, WDEntity],
        wdclasses: Mapping[str, WDClass],
        wdprops: Mapping[str, WDProperty],
        wd_numprop_stats: Mapping[str, WDQuantityPropertyStats],
        disable_rules: Set[str] = None,
        sim_fn: Optional[Callable[[str, str], float]] = None,
        cache_dir: Optional[str] = None,
        enable_logging: bool = False,
    ):
        self.wdclasses = wdclasses
        self.wdprops = wdprops
        self.wd_numprop_stats = wd_numprop_stats
        self.qnodes = qnodes
        self.temp_dir = f"/tmp/psl-python-{str(uuid.uuid4()).replace('-', '')}"
        self.sim_fn = sim_fn
        # use this cache dir to catch the extraction result
        self.cache_dir = Path(cache_dir) if cache_dir is not None else cache_dir

        # blacklist some rules if they do not apply to particular domain (synthetic dataset)
        all_rules = {
            self.LinkNegPrior,
            self.TypeNegPrior,
            self.LinkNegParentPropPrior,
            self.FreqLinkOverRow,
            self.FreqLinkOverEntRow,
            self.FreqLinkOverPosLink,
            self.FreqLinkUnmatchOverEntRow,
            self.FreqLinkUnmatchOverPossibleLink,
            self.LinkHeaderSimilarity,
            self.LinkDataTypeMismatch,
            self.FreqTypeOverRow,
            self.LinkNotFuncDep,
        }
        self.disable_rules = set(disable_rules) if disable_rules is not None else set()
        assert all(r in all_rules for r in self.disable_rules)

        self.enable_logging = enable_logging

        # load the model
        self.get_model()

    def get_model(self):
        """Construct the PSL model"""
        model = Model("st-solver")

        # ##################################################################
        # add predicates
        self.link_pos_feats = [
            r
            for r in [
                self.FreqLinkOverRow,
                self.FreqLinkOverEntRow,
                self.FreqLinkOverPosLink,
                self.LinkHeaderSimilarity,
            ]
            if r not in self.disable_rules
        ]
        self.link_neg_feats = [
            r
            for r in [
                self.FreqLinkUnmatchOverEntRow,
                self.FreqLinkUnmatchOverPossibleLink,
                self.LinkDataTypeMismatch,
            ]
            if r not in self.disable_rules
        ]
        self.link_structure_feats = [
            r for r in [self.LinkNotFuncDep] if r not in self.disable_rules
        ]
        self.link_all_feats = (
            self.link_pos_feats + self.link_neg_feats + self.link_structure_feats
        )
        self.type_pos_feats = [
            r
            for r in [self.FreqTypeInheritOverRow, self.TypeHeaderSimilarity]
            if r not in self.disable_rules
        ]
        self.type_all_feats = self.type_pos_feats

        model.add_predicate(Predicate("CanRel", closed=True, size=3))
        model.add_predicate(Predicate("CanType", closed=True, size=2))
        model.add_predicate(Predicate("Rel", closed=False, size=3))
        model.add_predicate(Predicate("Type", closed=False, size=2))
        model.add_predicate(Predicate("SubProp", closed=True, size=2))
        model.add_predicate(Predicate("SubType", closed=True, size=2))
        model.add_predicate(Predicate("HasSubType", closed=True, size=2))
        model.add_predicate(Predicate("NotRange", closed=True, size=2))
        model.add_predicate(Predicate("NotStatement", closed=True, size=1))
        model.add_predicate(Predicate("Statement", closed=True, size=1))

        for feat in self.link_all_feats:
            model.add_predicate(Predicate(f"RelFeature_{feat}", closed=True, size=3))
        for feat in self.type_all_feats:
            model.add_predicate(Predicate(f"TypeFeature_{feat}", closed=True, size=2))

        # ##################################################################
        # add rules
        feat_weights = {
            self.LinkNegPrior: 1,
            self.TypeNegPrior: 1,
            self.LinkNegParentPropPrior: 0.1,
            self.TypeNegParentPrior: 0.1,
            self.TypeHeaderSimilarity: 0.1,
            self.CascadingError: 2,
            self.TypeMustInPropRange: 2,
            self.FreqLinkOverRow: 2,
            self.FreqLinkOverEntRow: 2,
            self.FreqLinkOverPosLink: 2,
            self.FreqLinkUnmatchOverEntRow: 2,
            self.FreqLinkUnmatchOverPossibleLink: 2,
            self.LinkHeaderSimilarity: 2,
            self.LinkDataTypeMismatch: 2,
            self.FreqTypeOverRow: 2,
            self.FreqTypeInheritOverRow: 2,
            self.LinkNotFuncDep: 100,
        }
        self.rules = {}

        # rules apply for only link
        for feat in self.link_pos_feats:
            self.rules[feat] = Rule(
                f"CanRel(N1, N2, P) & RelFeature_{feat}(N1, N2, P) -> Rel(N1, N2, P)",
                weighted=True,
                squared=True,
                weight=feat_weights[feat],
            )
        for feat in self.link_neg_feats:
            self.rules[feat] = Rule(
                f"CanRel(N1, N2, P) & RelFeature_{feat}(N1, N2, P) -> ~Rel(N1, N2, P)",
                weighted=True,
                weight=feat_weights[feat],
                squared=True,
            )

        self.rules[self.LinkNotFuncDep] = Rule(
            f"CanRel(N1, N2, P) & CanRel(N2, N3, P2) & NotStatement(N2) & Rel(N1, N2, P) & RelFeature_{self.LinkNotFuncDep}(N2, N3, P2) -> ~Rel(N2, N3, P2)",
            weighted=True,
            weight=feat_weights[self.LinkNotFuncDep],
            squared=True,
        )

        # give a small negative weight to the parent property if there is a more specific child property
        # some how use CanRel is a much better option than Rel, perhaps this reflect how the parameters is set
        self.rules[self.LinkNegParentPropPrior] = [
            Rule(
                f"CanRel(N1, S, P) & Statement(S) & CanRel(S, N2, P) & CanRel(N1, S2, PP) & Statement(S2) & CanRel(S2, N2, PP) & SubProp(P, PP) -> ~Rel(N1, S2, PP)",
                weighted=True,
                weight=feat_weights[self.LinkNegParentPropPrior],
                squared=True,
            ),
            Rule(
                f"CanRel(N1, S, P) & Statement(S) & CanRel(S, N2, P) & CanRel(N1, S2, PP) & Statement(S2) & CanRel(S2, N2, PP) & SubProp(P, PP) -> ~Rel(S2, N2, PP)",
                weighted=True,
                weight=feat_weights[self.LinkNegParentPropPrior],
                squared=True,
            ),
        ]

        # give a small negative weight to the parent type
        self.rules[self.TypeNegParentPrior] = [
            Rule(
                # f"CanType(N, T1) & CanType(N, T2) & SubType(T1, T2) -> ~Type(N, T2)",
                "CanType(N, T) & HasSubType(N, T) -> ~Type(N, T)",
                weighted=True,
                weight=feat_weights[self.TypeNegParentPrior],
                squared=True,
            )
        ]

        # default negative prior
        self.rules[self.LinkNegPrior] = Rule(
            "~Rel(N1, N2, P)",
            weighted=True,
            weight=feat_weights[self.LinkNegPrior],
            squared=True,
        )
        self.rules[self.TypeNegPrior] = Rule(
            "~Type(N, P)",
            weighted=True,
            weight=feat_weights[self.TypeNegPrior],
            squared=True,
        )
        self.rules[self.CascadingError] = [
            Rule(
                f"CanRel(N0, S, P) & Statement(S) & CanRel(S, N1, P) & CanRel(S, N2, Q) & N1 != N2 & ~Rel(S, N1, P) -> ~Rel(S, N2, Q)",
                weighted=True,
                weight=feat_weights[self.CascadingError],
                squared=True,
            ),
            Rule(
                f"CanRel(N0, S, P) & Statement(S) & CanRel(S, N1, P) & ~Rel(S, N1, P) -> ~Rel(N0, S, P)",
                weighted=True,
                weight=feat_weights[self.CascadingError],
                squared=True,
            ),
            Rule(
                f"CanRel(N0, S, P) & Statement(S) & CanRel(S, N1, P) & ~Rel(N0, S, P) -> ~Rel(S, N1, P)",
                weighted=True,
                weight=feat_weights[self.CascadingError],
                squared=True,
            ),
        ]

        # self.rules[self.TypeMustInPropRange] = Rule(
        #     "CanRel(S, N1, P) & Statement(S) & CanType(N1, T) & NotRange(P, T) & Rel(S, N1, P) -> ~Type(N1, T)",
        #     weighted=True, weight=feat_weights[self.TypeMustInPropRange], squared=True)
        for feat in self.type_pos_feats:
            self.rules[feat] = Rule(
                f"CanType(N, T) & TypeFeature_{feat}(N, T) -> Type(N, T)",
                weighted=True,
                weight=feat_weights[feat],
                squared=True,
            )

        for rule_id, rules in self.rules.items():
            if rule_id in self.disable_rules:
                continue
            if isinstance(rules, list):
                for rule in rules:
                    model.add_rule(rule)
            else:
                model.add_rule(rules)

        # print rules (for debugging)
        # print("=" * 10, "Rules")
        # for rule_id, rules in self.rules.items():
        #     if rule_id in self.disable_rules:
        #         continue
        #     if isinstance(rules, list):
        #         for rule in rules:
        #             print(f"{rule_id}: {rule._rule_body}")
        #     else:
        #         print(f"{rule_id}: {rules._rule_body}")
        # print("=" * 10)

        # set the model and done
        self.model = model

    def infer(self, data: Dict[str, list]):
        """Run inference and get back the result.
        Note: Check the model to see the list of predicate we need to pass the data to.
        """
        preds = {x.upper() for x in data.keys()}
        miss_preds = [
            p
            for p in self.model.get_predicates()
            if p not in preds and p not in {"REL", "TYPE"}
        ]
        if len(miss_preds) > 0:
            raise Exception(
                f"Data in all predicates must be set. Missing {','.join(miss_preds)} predicates"
            )
        # TODO: fix me! temporary allow cantype to be empty
        can_be_empty_predicates = {
            "SubProp",
            f"RelFeature_{self.LinkDataTypeMismatch}",
            "NotRange",
            "CanType",
            f"TypeFeature_{self.FreqTypeOverRow}",
            f"TypeFeature_{self.FreqTypeInheritOverRow}",
            f"TypeFeature_{self.TypeHeaderSimilarity}",
            f"RelFeature_{self.LinkNotFuncDep}",
        }
        RelPredicate = (
            self.model.get_predicate("Rel")
            .clear_data()
            .add_data(Partition.TARGETS, data["CanRel"])
        )
        if len(data["CanType"]) == 0:
            TypePredicate = self.model.get_predicate("Type").clear_data()
        else:
            TypePredicate = (
                self.model.get_predicate("Type")
                .clear_data()
                .add_data(Partition.TARGETS, data["CanType"])
            )
        for pred, pred_data in data.items():
            if pred in can_be_empty_predicates and len(pred_data) == 0:
                self.model.get_predicate(pred).clear_data()
            else:
                self.model.get_predicate(pred).clear_data().add_data(
                    Partition.OBSERVATIONS, pred_data
                )

        infer_resp = self.model.infer(
            logger=None if self.enable_logging else False,
            additional_cli_optons=["--h2path", os.path.join(self.temp_dir, "h2")],
            temp_dir=str(self.temp_dir),
            cleanup_temp=False,
        )
        return {
            "links": {
                (r[0], r[1], r[2]): r["truth"]
                for ri, r in infer_resp[RelPredicate].iterrows()
            },
            "types": {
                (r[0], r[1]): r["truth"]
                for ri, r in infer_resp[TypePredicate].iterrows()
            },
        }

    def run(self, table: LinkedTable, dg: DGGraph, cg: CGGraph):
        inferred_result = self.solve(table, cg, dg)
        cpa = inferred_result["links"]
        cta = inferred_result["types"]
        cta = {int(ci.replace("column-", "")): classes for ci, classes in cta.items()}
        return cpa, cta

    def solve(self, table: LinkedTable, sg: CGGraph, dg: DGGraph):
        if sg.num_edges() == 0:
            return {"links": {}, "types": {}}

        data, idmap = self.extract_predicate_data(
            [(table, sg, dg)], is_parallel=False, show_progress=False
        )

        def unpack_mappedid(key):
            table_id, real_key = idmap.im(key)
            assert table_id == table.id
            return real_key

        infer_resp = self.infer(data)
        infer_resp_types = {}
        for (uid, classid), prob in infer_resp["types"].items():
            _ouid = unpack_mappedid(uid)
            if _ouid not in infer_resp_types:
                infer_resp_types[_ouid] = {}
            infer_resp_types[_ouid][classid] = prob

        infer_resp["links"] = {
            (unpack_mappedid(uid), unpack_mappedid(vid), eid): prob
            for (uid, vid, eid), prob in infer_resp["links"].items()
        }
        infer_resp["types"] = infer_resp_types

        # set the prob. for debugging purpose
        link_feat_extractor = LinkFeatureExtraction(
            table, sg, dg, self.qnodes, self.wdprops, self.wd_numprop_stats, self.sim_fn
        )
        link_feat_extractor.add_debug_info(link_feat_extractor.extract_features())
        for k, prob in infer_resp["links"].items():
            sg.get_edge_between_nodes(*k).features["prob"] = prob

        return infer_resp

    def train_setup(self, inputs: List[Tuple[LinkedTable, CGGraph, DGGraph, dict]]):
        data, idmap = self.extract_predicate_data(
            inputs=[(x[0], x[1], x[2]) for x in inputs], features=[x[3] for x in inputs]
        )

        assert (
            len(self.model.get_predicates()) == len(data) + 2
        ), "Data in all predicates must be set"
        can_be_empty_predicates = {"SubProp", f"RelFeature_{self.LinkDataTypeMismatch}"}
        self.model.get_predicate("Rel").clear_data().add_data(
            Partition.TARGETS, data["CanRel"]
        )
        self.model.get_predicate("Type").clear_data().add_data(
            Partition.TARGETS, data["CanType"]
        )
        for pred, pred_data in data.items():
            if pred in can_be_empty_predicates and len(pred_data) == 0:
                self.model.get_predicate(pred).clear_data()
            else:
                self.model.get_predicate(pred).clear_data().add_data(
                    Partition.OBSERVATIONS, pred_data
                )

        # clean up the temporary directory first. this code is takening from the PSL model.infer method
        if Path(self.temp_dir).exists():
            self.model._cleanup_temp(str(self.temp_dir))
        # this write out the data file and rules_file
        logger, temp_dir, data_file_path, rules_file_path = self.model._prep_run(
            logger=None if self.enable_logging else False, temp_dir=str(self.temp_dir)
        )
        cli_options = []
        cli_options.append("--infer")
        inferred_dir = os.path.join(temp_dir, Model.CLI_INFERRED_OUTPUT_DIR)
        cli_options.append("--output")
        cli_options.append(inferred_dir)

        self.train_args = dict(
            data_file_path=data_file_path,
            rules_file_path=rules_file_path,
            cli_options=cli_options,
            psl_config={},
            jvm_options=[],
            logger=logger,
            inferred_dir=inferred_dir,
            temp_dir=temp_dir,
        )
        return idmap

    def train_set_parameters(self, rules: Dict[str, float]):
        for rule_id, rule_weight in rules.items():
            if isinstance(self.rules[rule_id], list):
                for rule in self.rules[rule_id]:
                    rule.set_weight(rule_weight)
            else:
                self.rules[rule_id].set_weight(rule_weight)
        self.model._write_rules(self.train_args["temp_dir"])

    def train_eval(self, idmap: IDMap) -> Dict[str, Dict[Tuple[str, str, str], float]]:
        table_infer_resp = {}
        RelPredicate = self.model.get_predicate("Rel")

        self.model._run_psl(
            self.train_args["data_file_path"],
            self.train_args["rules_file_path"],
            self.train_args["cli_options"],
            self.train_args["psl_config"],
            self.train_args["jvm_options"],
            self.train_args["logger"],
        )
        infer_resp = self.model._collect_inference_results(
            self.train_args["inferred_dir"]
        )

        for ri, r in infer_resp[RelPredicate].iterrows():
            uid, vid, eid, prob = r[0], r[1], r[2], r["truth"]
            table_id, uid = idmap.im(uid)
            assert table_id == idmap.im(vid)[0]
            vid = idmap.im(vid)[1]

            if table_id not in table_infer_resp:
                table_infer_resp[table_id] = {}
            table_infer_resp[table_id][uid, vid, eid] = prob
        return table_infer_resp

    def extract_predicate_data(
        self,
        inputs: List[Tuple[LinkedTable, CGGraph, DGGraph]],
        features: Optional[List[dict]] = None,
        idmap=None,
        is_parallel: Union[str, bool] = "auto",
        show_progress: bool = False,
    ):
        """Extract predicates data from the list of tables, sgs, and dgs. If the features do not provided, it will
        be created automatically.

        If parallel is 'auto' we run parallel when length of inputs > 1
        """
        global global_objects
        if idmap is None:
            idmap = IDMap()

        if features is None:
            is_parallel = len(inputs) > 1 if is_parallel == "auto" else is_parallel
            global_objects["wdprops"] = self.wdprops
            global_objects["wdclasses"] = self.wdclasses
            global_objects["qnodes"] = self.qnodes
            global_objects["wd_numprop_stats"] = self.wd_numprop_stats
            global_objects["sim_fn"] = self.sim_fn

            if is_parallel:
                global_objects["cache_dir"] = self.cache_dir
            else:
                global_objects["cache_dir"] = None
            pp = Parallel()
            features = pp.map(
                PSLInference._extract_features_wrapper,
                list(enumerate(inputs)),
                show_progress=show_progress,
                progress_desc="psl: extract features",
                is_parallel=is_parallel,
            )
        assert features is not None
        data = {
            "CanRel": [],
            "CanType": [],
            "NotStatement": [],
            "Statement": [],
            "SubProp": set(),
            "SubType": set(),
            "HasSubType": set(),
            "NotRange": set(),
        }
        for feat in self.link_all_feats:
            data[f"RelFeature_{feat}"] = []
        for feat in self.type_all_feats:
            data[f"TypeFeature_{feat}"] = []
        for (table, sg, dg), result in zip(inputs, features):
            table_id = table.id
            idmap.add_keys([(table_id, u.id) for u in sg.iter_nodes()])
            data["CanRel"] += [
                (
                    idmap.m((table_id, edge.source)),
                    idmap.m((table_id, edge.target)),
                    edge.predicate,
                )
                for edge in sg.iter_edges()
            ]
            data[f"CanType"] += [
                (idmap.m((table_id, uid)), classid)
                for uid, classid in result["type_feats"][TypeFeatureExtraction.Freq]
            ]
            data["NotStatement"] += [
                (idmap.m((table_id, u.id)),)
                for u in sg.iter_nodes()
                if not isinstance(u, CGStatementNode)
            ]
            data["Statement"] += [
                (idmap.m((table_id, u.id)),)
                for u in sg.iter_nodes()
                if isinstance(u, CGStatementNode)
            ]

            props = {edge.predicate for edge in sg.iter_edges()}
            for p in props:
                for pp in props:
                    if p != pp and pp in self.wdprops[p].parents_closure:
                        data["SubProp"].add((p, pp))

            for uid, class_ids in result["type_feats"]["_column_to_types"].items():
                class_ids = set(class_ids)
                for class_id in class_ids:
                    for parent_class_id in self.wdclasses[class_id].parents:
                        if parent_class_id in class_ids:
                            data["SubType"].add((class_id, parent_class_id))
                            data["HasSubType"].add(
                                (idmap.m((table_id, uid)), parent_class_id)
                            )

            for uid, class_ids in result["type_feats"]["_column_to_types"].items():
                # find list of incoming edges
                incoming_props = {edge.predicate for edge in sg.in_edges(uid)}
                incoming_props = [self.wdprops[eid] for eid in incoming_props]
                for class_id in class_ids:
                    if class_id not in self.wdclasses:
                        continue
                    parents_closure = self.wdclasses[class_id].parents_closure
                    for p in incoming_props:
                        if len(p.subjects) > 0 and not any(
                            class_id == subj or subj in parents_closure
                            for subj in p.subjects
                        ):
                            data["NotRange"].add((p.id, class_id))

            for feat in self.link_all_feats:
                data[f"RelFeature_{feat}"] += [
                    (idmap.m((table_id, uid)), idmap.m((table_id, vid)), eid, prob)
                    for (uid, vid, eid), prob in result["link_feats"][feat].items()
                ]
            for feat in self.type_all_feats:
                data[f"TypeFeature_{feat}"] += [
                    (idmap.m((table_id, uid)), classid, prob)
                    for (uid, classid), prob in result["type_feats"][feat].items()
                ]
        return data, idmap

    @staticmethod
    def _update_global_objects(**kwargs):
        global global_objects
        for k, v in kwargs.items():
            global_objects[k] = v

    @staticmethod
    def _extract_features_wrapper(
        args: Tuple[int, Tuple[LinkedTable, CGGraph, DGGraph]]
    ):
        """Wrap the function that need to pass some big objects through a global objects"""
        global global_objects
        wdprops: Mapping[str, WDProperty] = global_objects["wdprops"]
        wdclasses: Mapping[str, WDClass] = global_objects["wdclasses"]
        wd_numprop_stats = global_objects["wd_numprop_stats"]
        qnodes: Mapping[str, WDEntity] = global_objects["qnodes"]
        cache_dir = global_objects["cache_dir"]
        sim_fn = global_objects["sim_fn"]

        index, (table, sg, dg) = args

        if cache_dir is not None:
            (cache_dir / "features").mkdir(exist_ok=True, parents=True)
            filename = table.get_friendly_fs_id()
            filepath = cache_dir / "features" / f"a{index:03d}_{filename}.pkl"
            if filepath.exists():
                return M.deserialize_pkl(filepath)
        else:
            filepath = None

        start = time.time()
        link_feat_extractor = LinkFeatureExtraction(
            table, sg, dg, qnodes, wdprops, wd_numprop_stats, sim_fn
        )
        type_feat_extractor = TypeFeatureExtraction(
            table, sg, dg, qnodes, wdclasses, wdprops, wd_numprop_stats, sim_fn
        )

        link_feats = link_feat_extractor.extract_features()
        type_feats = type_feat_extractor.extract_features()

        exectime = time.time() - start
        props = {edge.predicate for edge in sg.iter_edges()}
        sub_props = set()
        for p in props:
            for pp in props:
                if p != pp and pp in wdprops[p].parents_closure:
                    sub_props.add((p, pp))
        prop_ranges = set()

        cache_content = {
            "link_feats": link_feats,
            "type_feats": type_feats,
            "sub_props": sub_props,
            "exec_time": exectime,
        }
        if filepath is not None:
            M.serialize_pkl(cache_content, filepath)
        return cache_content



from collections import defaultdict
from grams.algorithm.data_graph.dg_config import DGConfigs
import networkx as nx
from typing import Dict, List, Optional, Tuple, TypedDict, Union, cast

from functools import cmp_to_key
from grams.algorithm.data_graph.dg_graph import (
    CellNode,
    DGEdge,
    DGGraph,
    DGNode,
    EdgeFlowSource,
    EdgeFlowTarget,
    EntityValueNode,
    FlowProvenance,
    LinkGenMethod,
    LiteralValueNode,
    StatementNode,
)
from graph.retworkx import digraph_all_simple_paths


class DGPruning:
    NxDGEdgeAttr = TypedDict("NxDGEdgeAttr", data=DGEdge)
    NxDGEdge = Tuple[str, str, str, NxDGEdgeAttr]

    def __init__(self, dg: DGGraph):
        self.dg = dg

    def prune_hidden_entities(self):
        """Prune redundant KG entities, which added to the graph via KG discovering and from the context.

        **Step 1:**
        Let:
        - n be an entity node in DG.
        - v is a node connected from n via a property: LEG2: n -> p -> s -> p' -> v, and s does not have other property/qualifier rather than p'

        We made the following heuristics:
        * If there is no other node connect to n, then n is a root node and is from the context. We should not
        prune this node, so just skip it.
        * For all another node ui in U that connects to n via the path: LEG1: ui -> pi -> s' -> pi' -> n, if there is always a better
        path LEG* between ui and v, then we can remove the path LEG2. U contains nodes in cells or context, otherwise
        ui will be an entity to entity that won't be in the final model anyway.
        Note: LEG* is better than LEG2 when it's shorter, also from wikidata link or if not, it must have better match confidence (i.e., better provenance)

        **Step 2:** (?) this is questionable -- controlled by PRUNE_SINGLE_LEAF_ENT flag
        Let n' be an entity node in DG that do not link to other nodes (v doesn't exist).
        We have the following heuristics:
        * If there is no other node connect to it, this is a standable node and should be removed
        * For any node ui that connects to n via the path: LEG1: ui -> pi -> s' -> pi' -> n. If s' doesn't have other properties/qualifiers,
        then we can remove LEG1.

        **Step 3:**
        Let n be an entity/literal node in DG (not from cells & context)
        * If there is only one entity node u that connects to n via the path: LEG1: u -> p -> s -> p' -> n, then we can remove n. If that renders the
        statement s without value, we will remove s as well.

        **Step 3:**
        Finally, if a node is standalone, we should remove it.
        """
        # step 1: prune the second leg paths
        legprime: Dict[Tuple[str, str], Optional[FlowProvenance]] = {}
        rm_legs: List[Tuple[str, EdgeFlowSource, EdgeFlowTarget]] = []
        for n in self.dg.iter_nodes():
            if not isinstance(n, EntityValueNode):
                continue

            if self.dg.in_degree(n.id) == 0:
                # no other node connect to it
                continue

            # get list of grandpa ui (only cells or values in the context), along with their paths to node n.
            grandpa = set()
            for gp in self.iter_grand_parents(n.id):
                if isinstance(gp, CellNode) or (
                    isinstance(gp, (EntityValueNode, LiteralValueNode))
                    and gp.is_context
                ):
                    grandpa.add(gp.id)

            # for _, sid, ns_eid in self.dg.out_edges(nid, keys=True):
            #     stmt: StatementNode = self.dg.nodes[sid]["data"]
            for ns_edge in self.dg.out_edges(n.id):
                stmt = self.dg.get_node(ns_edge.target)
                assert isinstance(stmt, StatementNode)
                stmt_outedges = self.dg.out_edges(stmt.id)
                if len(stmt_outedges) > 1:
                    # this stmt has other qualifiers, so it's not what we are looking for
                    continue

                for sv_outedge in stmt_outedges:
                    v = self.dg.get_node(sv_outedge.target)
                    # got leg 2, now looks for all incoming
                    leg2 = (
                        EdgeFlowSource(n.id, ns_edge.predicate),
                        EdgeFlowTarget(v.id, sv_outedge.predicate),
                    )
                    if not stmt.has_flow(*leg2):
                        continue
                    leg2_provenance = stmt.get_provenance(*leg2)

                    has_better_paths = True
                    for gpid in grandpa:
                        if (gpid, v.id) not in legprime:
                            paths = [
                                (
                                    cast(
                                        StatementNode, self.dg.get_node(path[0].target)
                                    ),
                                    EdgeFlowSource(path[0].source, path[0].predicate),
                                    EdgeFlowTarget(path[1].target, path[1].predicate),
                                )
                                for path in digraph_all_simple_paths(
                                    self.dg,
                                    gpid,
                                    v.id,
                                    cutoff=2,
                                )
                            ]
                            provs = [
                                prov
                                for s, sf, tf in paths
                                if s.has_flow(sf, tf)
                                for prov in s.get_provenance(sf, tf)
                            ]
                            if len(provs) == 0:
                                legprime[gpid, v.id] = None
                            else:
                                legprime[gpid, v.id] = max(
                                    provs,
                                    key=cmp_to_key(
                                        self.specific_pruning_provenance_cmp
                                    ),
                                )
                        best_prov = legprime[gpid, v.id]
                        if (
                            best_prov is None
                            or max(
                                self.specific_pruning_provenance_cmp(
                                    best_prov, leg2_prov
                                )
                                for leg2_prov in leg2_provenance
                            )
                            < 0
                        ):
                            # no better path
                            has_better_paths = False
                            break

                    if has_better_paths:
                        rm_legs.append((stmt.id, leg2[0], leg2[1]))

        # logger.info("#legs: {}", len(rm_legs))
        self.remove_flow(rm_legs)
        # logger.info("# 0-indegree: {}", sum(self.dg.in_degree(uid) == 0 for uid in self.dg.nodes))
        # logger.info("# 0-outdegree: {}", sum(self.dg.out_degree(uid) == 0 for uid in self.dg.nodes))
        # logger.info("# 0-standalone: {}",
        #             sum(self.dg.out_degree(uid) + self.dg.in_degree(uid) == 0 for uid in self.dg.nodes))

        # step 2: prune the first leg paths (temporary disable)
        if DGConfigs.PRUNE_SINGLE_LEAF_ENT:
            rm_legs: List[Tuple[str, EdgeFlowSource, EdgeFlowTarget]] = []
            for n in self.dg.iter_nodes():
                if not isinstance(n, EntityValueNode) or self.dg.out_degree(n.id) > 0:
                    continue

                for sn_edge in self.dg.in_edges(n.id):
                    if self.dg.out_degree(sn_edge.source) == 1:
                        # stmt does not have other property/qualifier
                        target_flow = EdgeFlowTarget(n.id, sn_edge.predicate)
                        stmt = self.dg.get_node(sn_edge.source)
                        assert isinstance(stmt, StatementNode)
                        for source_flow, _ in stmt.iter_source_flow(target_flow):
                            rm_legs.append((sn_edge.source, source_flow, target_flow))
            self.remove_flow(rm_legs)

        rm_legs: List[Tuple[str, EdgeFlowSource, EdgeFlowTarget]] = []
        # step 3
        for n in self.dg.iter_nodes():
            if (
                not isinstance(n, (EntityValueNode, LiteralValueNode))
                or self.dg.out_degree(n.id) > 0
                or n.is_context
            ):
                continue
            # for sid, _, sn_eid, sn_edata in self.dg.in_edges(nid, data=True, keys=True):
            for sn_edge in self.dg.in_edges(n.id):
                # for uid, _, us_eid in self.dg.in_edges(sn_edge.source):
                for us_edge in self.dg.in_edges(sn_edge.source):
                    if isinstance(self.dg.get_node(us_edge.source), EntityValueNode):
                        # two consecutive entity nodes, we can remove this link
                        stmt = self.dg.get_node(sn_edge.source)
                        assert isinstance(stmt, StatementNode)
                        if us_edge.predicate == sn_edge.predicate:
                            # the link we are going to remove is the statement value, so we should remove the statement
                            for source_flow, target_flow in stmt.flow:
                                rm_legs.append((stmt.id, source_flow, target_flow))
                        else:
                            target_flow = EdgeFlowTarget(n.id, sn_edge.predicate)
                            for source_flow, _ in stmt.iter_source_flow(target_flow):
                                rm_legs.append((stmt.id, source_flow, target_flow))

        # logger.info("#legs: {}", len(rm_legs))
        self.remove_flow(rm_legs)
        # logger.info("# 0-indegree: {}", sum(self.dg.in_degree(uid) == 0 for uid in self.dg.nodes))
        # logger.info("# 0-outdegree: {}", sum(self.dg.out_degree(uid) == 0 for uid in self.dg.nodes))
        # logger.info("# 0-standalone: {}",
        #             sum(self.dg.out_degree(uid) + self.dg.in_degree(uid) == 0 for uid in self.dg.nodes))
        self.prune_disconnected_nodes()

    def prune_disconnected_nodes(self):
        """This function prune out disconnected nodes that are:
        1. nodes without incoming edges and outgoing edges
        2. statement nodes with no incoming edges or no outgoing edges

        Returns
        -------
        """
        rm_nodes = set()
        for u in self.dg.iter_nodes():
            if isinstance(u, EntityValueNode):
                if self.dg.in_degree(u.id) == 0 and self.dg.out_degree(u.id) == 0:
                    rm_nodes.add(u.id)
            elif isinstance(u, StatementNode):
                if self.dg.in_degree(u.id) == 0 or self.dg.out_degree(u.id) == 0:
                    rm_nodes.add(u.id)
        for uid in rm_nodes:
            self.dg.remove_node(uid)

    def remove_flow(self, flows: List[Tuple[str, EdgeFlowSource, EdgeFlowTarget]]):
        for sid, source_flow, target_flow in flows:
            stmt = cast(StatementNode, self.dg.get_node(sid))
            stmt.untrack_flow(source_flow, target_flow)
            if not stmt.has_source_flow(source_flow):
                self.dg.remove_edge_between_nodes(
                    source_flow.source_id, sid, source_flow.edge_id
                )
            if not stmt.has_target_flow(target_flow):
                self.dg.remove_edge_between_nodes(
                    sid, target_flow.target_id, target_flow.edge_id
                )

    def specific_pruning_provenance_cmp(
        self, prov0: FlowProvenance, prov1: FlowProvenance
    ) -> int:
        # compare provenance, this function only accept
        if prov0.gen_method == LinkGenMethod.FromWikidataLink:
            # always favour from wikidata link
            return 1
        if prov1.gen_method == LinkGenMethod.FromWikidataLink:
            return -1
        # assert prov0.gen_method == prov1.gen_method and prov0.gen_method_arg == prov1.gen_method_arg
        # do not need to check if the two gen method and args are equal, as even if we select the incorrect one
        # we only truncate when the other leg worst than it
        return prov0.prob - prov1.prob  # type: ignore

    def iter_grand_parents(self, nid: str):
        for parent in self.dg.predecessors(nid):
            for grand_parent in self.dg.predecessors(parent.id):
                yield grand_parent

    # def out_edges(self, uid: str) -> Dict[str, List[DGEdge]]:
    #     label2edges = defaultdict(list)
    #     for _, vid, eid, edata in self.dg.out_edges(uid, data=True, keys=True):
    #         label2edges[eid].append(edata["data"])
    #     return label2edges



from typing import Dict, List, Tuple, TypedDict
from grams.algorithm.candidate_graph.cg_graph import (
    CGColumnNode,
    CGEdge,
    CGEntityValueNode,
    CGGraph,
    CGLiteralValueNode,
    CGStatementNode,
)
from operator import itemgetter


_AddContextPathType = TypedDict(
    "_Path", {"path": Tuple[CGEdge, CGEdge], "score": float}
)


def ensure_valid_statements(
    original_cg: CGGraph,
    pruned_cg: CGGraph,
    create_if_not_exists: bool = False,
    safe_creation: bool = True,
):
    """Ensure that the pruned graph has valid statements, which have statement property.

    NOTE: this function may modify the pruned graph.

    Args:
        original_cg: the original candidate graph
        pruned_cg: the pruned candidate graph -- must have the same id as the original cg
        create_if_not_exists: if True, create a new target of statement property if it does not exist
        safe_creation: if True, we check to make sure the new statement target is not in the graph
    """
    for s in pruned_cg.nodes():
        if not isinstance(s, CGStatementNode):
            continue

        (inedge,) = pruned_cg.in_edges(s.id)
        for outedge in original_cg.out_edges(s.id):
            if outedge.predicate == inedge.predicate:
                # there is a statement prop in the original candidate graph
                # if it is not in the tree, we need to add it back
                if pruned_cg.has_edge_between_nodes(
                    outedge.source, outedge.target, outedge.predicate
                ):
                    continue

                if not create_if_not_exists:
                    raise Exception(
                        f"The property of the statement {s.id} is not in the graph"
                    )

                if safe_creation:
                    assert not pruned_cg.has_node(
                        outedge.target
                    ), "The target should not in the grpah as we do not have the link"
                    target = original_cg.get_node(outedge.target).clone()
                    assert isinstance(
                        target, (CGEntityValueNode, CGLiteralValueNode)
                    ), "The only reason why we don't have statement value is it is an entity/literal"
                    pruned_cg.add_node(target)
                elif not pruned_cg.has_node(outedge.target):
                    target = original_cg.get_node(outedge.target).clone()
                    pruned_cg.add_node(target)

                pruned_cg.add_edge(outedge.clone())


def add_context(
    origin_cg: CGGraph,
    pruned_cg: CGGraph,
    edge_probs: Dict[Tuple[str, str, str], float],
):
    """Add back the context node or entity that are appeared in the original graph but not in the pruned graph.

    Note: this function may modify the pruned graph.

    Args:
        origin_cg: the original candidate graph
        pruned_cg: the pruned candidate graph -- must have the same id as the original cg
        edge_probs: containing edges' probabilities
    """
    for v in origin_cg.iter_nodes():
        if (
            not isinstance(v, (CGEntityValueNode, CGLiteralValueNode))
            or not v.is_in_context
            or pruned_cg.has_node(v.id)
        ):
            continue

        # find the paths that connect the vid to the tree and select the one with highest score and do not create cycle
        paths: List[_AddContextPathType] = []
        for sv_edge in origin_cg.in_edges(v.id):
            if sv_edge.predicate == "P31":  # ignore instanceof
                continue
            for us_edge in origin_cg.in_edges(sv_edge.source):
                if not pruned_cg.has_node(us_edge.source):
                    continue
                paths.append(
                    {
                        "path": (
                            us_edge,
                            sv_edge,
                        ),
                        "score": edge_probs[
                            (us_edge.source, us_edge.target, us_edge.predicate)
                        ]
                        + edge_probs[
                            (sv_edge.source, sv_edge.target, sv_edge.predicate)
                        ],
                    }
                )

        paths = sorted(paths, key=itemgetter("score"), reverse=True)

        if len(paths) == 0:
            continue

        us_edge, sv_edge = paths[0]["path"]
        pruned_cg.add_node(v.clone())
        if not pruned_cg.has_node(sv_edge.source):
            s = origin_cg.get_node(sv_edge.source)
            pruned_cg.add_node(s.clone())
        assert not pruned_cg.has_edge_between_nodes(
            sv_edge.source, sv_edge.target, sv_edge.predicate
        )
        pruned_cg.add_edge(sv_edge.clone())

        if not pruned_cg.has_edge_between_nodes(
            us_edge.source, us_edge.target, us_edge.predicate
        ):
            pruned_cg.add_edge(us_edge.clone())



from typing import Tuple
from grams.algorithm.literal_matchers.text_parser import ParsedTextRepr
from kgdata.wikidata.models import WDValue
from grams.algorithm.literal_matchers.types import LiteralMatchKit


def globecoordinate_test(
    kgvalue: WDValue, value: ParsedTextRepr, kit: LiteralMatchKit
) -> Tuple[bool, float]:
    """Compare if the value in KG matches with value in the cell

    Args:
        kgvalue: the value in KG
        value: the value in the cell
        kit: objects that may be needed for matching

    Returns:
        a tuple of (whether it's matched, confidence)
    """
    return False, 0.0



import itertools
from enum import IntEnum
from typing import Dict, List, Mapping, Optional, Set, Tuple
from uuid import uuid4
from grams.algorithm.candidate_graph.cg_graph import (
    CGColumnNode,
    CGEntityValueNode,
    CGGraph,
    CGLiteralValueNode,
    CGStatementNode,
)

import networkx as nx
import numpy as np
import sm.misc as M
import sm.outputs as O

from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import WDEntity, WDClass, WDProperty
from rdflib import RDFS
from sm.evaluation import sm_metrics
from sm.misc import identity_func
from sm.misc.graph import viz_graph
from sm.outputs.semantic_model import ClassNode
from sm.namespaces import WikidataNamespace


class SMNodeType(IntEnum):
    Column = 0
    Class = 1
    Statement = 2
    Entity = 3
    Literal = 4


class WikidataSemanticModelHelper:
    ENTITY_ID = "Q35120"
    ENTITY_LABEL = "Entity (Q35120)"
    ID_PROPS = {str(RDFS.label)}

    def __init__(
        self,
        wdentities: Mapping[str, WDEntity],
        wdentity_labels: Mapping[str, str],
        wdclasses: Mapping[str, WDClass],
        wdprops: Mapping[str, WDProperty],
    ):
        self.wdentities = wdentities
        self.wdentity_labels = wdentity_labels
        self.wdclasses = wdclasses
        self.wdprops = wdprops
        self.wdns = WikidataNamespace.create()

    def norm_sm(self, sm: O.SemanticModel):
        """ "Normalize the semantic model with the following modifications:
        1. Add readable label to edge and class
        2. Convert direct link (without statement) to have statement except the id props.
        """
        new_sm = sm.deep_copy()
        wdns = self.wdns

        # update readable label
        for n in new_sm.iter_nodes():
            if isinstance(n, O.ClassNode):
                if wdns.is_abs_uri_qnode(n.abs_uri):
                    n.readable_label = self.get_qnode_label(
                        wdns.get_entity_id(n.abs_uri)
                    )
            elif isinstance(n, O.LiteralNode):
                if wdns.is_abs_uri_qnode(n.value):
                    n.readable_label = self.get_qnode_label(wdns.get_entity_id(n.value))
        for e in new_sm.iter_edges():
            if e.abs_uri not in self.ID_PROPS:
                e.readable_label = self.get_pnode_label(wdns.get_prop_id(e.abs_uri))

        # convert direct link
        for edge in list(new_sm.iter_edges()):
            if edge.abs_uri in self.ID_PROPS:
                continue
            source = new_sm.get_node(edge.source)
            target = new_sm.get_node(edge.target)

            if (
                not isinstance(source, O.ClassNode)
                or source.abs_uri != WikidataNamespace.STATEMENT_URI
            ) and (
                not isinstance(target, O.ClassNode)
                or target.abs_uri != WikidataNamespace.STATEMENT_URI
            ):
                # this is direct link, we replace its edge
                assert len(new_sm.get_edges_between_nodes(source.id, target.id)) == 1
                new_sm.remove_edges_between_nodes(source.id, target.id)
                stmt = O.ClassNode(
                    abs_uri=WikidataNamespace.STATEMENT_URI,
                    rel_uri=wdns.get_rel_uri(WikidataNamespace.STATEMENT_URI),
                )
                new_sm.add_node(stmt)

                new_sm.add_edge(
                    O.Edge(
                        source=edge.source,
                        target=stmt.id,
                        abs_uri=edge.abs_uri,
                        rel_uri=edge.rel_uri,
                        approximation=edge.approximation,
                        readable_label=edge.readable_label,
                    )
                )
                new_sm.add_edge(
                    O.Edge(
                        source=stmt.id,
                        target=edge.target,
                        abs_uri=edge.abs_uri,
                        rel_uri=edge.rel_uri,
                        approximation=edge.approximation,
                        readable_label=edge.readable_label,
                    )
                )
        return new_sm

    @staticmethod
    def minify_sm(sm: O.SemanticModel):
        """This is a reverse function of `norm_sm`:
        1. Remove an intermediate statement if it doesn't have any qualifiers
        """
        new_sm = sm.copy()

        for n in sm.iter_nodes():
            if (
                isinstance(n, O.ClassNode)
                and n.abs_uri == WikidataNamespace.STATEMENT_URI
            ):
                inedges = sm.in_edges(n.id)
                outedges = sm.out_edges(n.id)
                if len(outedges) == 1 and outedges[0].abs_uri == inedges[0].abs_uri:
                    # no qualifiers
                    new_sm.remove_node(n.id)
                    for inedge in inedges:
                        assert inedge.abs_uri == outedges[0].abs_uri
                        new_sm.add_edge(
                            O.Edge(
                                inedge.source,
                                outedges[0].target,
                                inedge.abs_uri,
                                inedge.rel_uri,
                                # just in case user misannotate to not include approximation in both links
                                inedge.approximation or outedges[0].approximation,
                                inedge.readable_label,
                            )
                        )
        return new_sm

    def create_sm(self, table: LinkedTable, cpa: CGGraph, cta: Dict[int, str]):
        """Create a semantic model from outputs of CPA and CTA tasks"""
        sm = O.SemanticModel()
        classmap = {}  # mapping from column to its class node
        wdns = self.wdns
        for cid, qnode_id in cta.items():
            dnode = O.DataNode(
                col_index=cid,
                label=table.table.columns[cid].name or "",
            )

            # somehow, they may end-up predict multiple classes, we need to select one
            if qnode_id.find(" ") != -1:
                qnode_id = qnode_id.split(" ")[0]
            curl = wdns.get_entity_abs_uri(qnode_id)

            try:
                cnode_label = self.get_qnode_label(qnode_id)
            except KeyError:
                cnode_label = wdns.get_entity_rel_uri(qnode_id)
            cnode = O.ClassNode(
                abs_uri=curl,
                rel_uri=wdns.get_entity_rel_uri(qnode_id),
                readable_label=cnode_label,
            )
            sm.add_node(dnode)
            sm.add_node(cnode)
            classmap[dnode.col_index] = cnode.id
            sm.add_edge(
                O.Edge(
                    source=cnode.id,
                    target=dnode.id,
                    abs_uri=str(RDFS.label),
                    rel_uri=wdns.get_rel_uri(RDFS.label),
                )
            )

        # do a final sweep to add subject columns that are not in CTA
        for unode in cpa.iter_nodes():
            if not isinstance(unode, CGColumnNode):
                continue
            outdegree: int = cpa.out_degree(unode.id)
            if outdegree > 0 and not sm.has_data_node(unode.column):
                # add data node to the graph and use the entity class (all instances belong to this class) to describe this data node
                dnode = O.DataNode(
                    col_index=unode.column,
                    label=table.table.columns[unode.column].name or "",
                )
                sm.add_node(dnode)

                curl = wdns.get_entity_abs_uri(self.ENTITY_ID)
                cnode_id = sm.add_node(
                    O.ClassNode(
                        abs_uri=curl,
                        rel_uri=wdns.get_entity_rel_uri(self.ENTITY_ID),
                        readable_label=self.ENTITY_LABEL,
                    )
                )
                classmap[dnode.col_index] = cnode_id

                sm.add_edge(
                    O.Edge(
                        source=cnode_id,
                        target=dnode.id,
                        abs_uri=str(RDFS.label),
                        rel_uri="rdfs:label",
                    )
                )

        # now add remaining edges and remember to use class node instead of data node
        cpa_idmap = {}
        for edge in cpa.edges():
            unode = cpa.get_node(edge.source)
            vnode = cpa.get_node(edge.target)

            if isinstance(unode, CGColumnNode):
                # outgoing edge is from a class node instead of a data node
                suid = classmap[unode.column]
                source = sm.get_node(suid)
            elif isinstance(unode, CGEntityValueNode):
                if unode.id not in cpa_idmap:
                    source = O.LiteralNode(
                        value=wdns.get_entity_abs_uri(unode.qnode_id),
                        readable_label=self.get_qnode_label(unode.qnode_id),
                        datatype=O.LiteralNodeDataType.Entity,
                        is_in_context=unode.qnode_id == table.context.page_entity_id,
                    )
                    cpa_idmap[unode.id] = sm.add_node(source)
                else:
                    source = sm.get_node(cpa_idmap[unode.id])
            else:
                assert isinstance(
                    unode, CGStatementNode
                ), "Outgoing edge can't not be from literal"
                if unode.id not in cpa_idmap:
                    # create a statement node
                    source = O.ClassNode(
                        abs_uri=wdns.STATEMENT_URI,
                        rel_uri=wdns.get_rel_uri(wdns.STATEMENT_URI),
                    )
                    cpa_idmap[unode.id] = sm.add_node(source)
                else:
                    source = sm.get_node(cpa_idmap[unode.id])

            if isinstance(vnode, CGColumnNode):
                if vnode.column in classmap:
                    target = sm.get_node(classmap[vnode.column])
                elif sm.has_data_node(vnode.column):
                    target = sm.get_data_node(vnode.column)
                elif vnode.id not in cpa_idmap:
                    target = O.DataNode(
                        col_index=vnode.column,
                        label=table.table.columns[vnode.column].name or "",
                    )
                    cpa_idmap[vnode.id] = sm.add_node(target)
                else:
                    target = sm.get_node(cpa_idmap[vnode.id])
            elif isinstance(vnode, CGEntityValueNode):
                if vnode.id not in cpa_idmap:
                    target = O.LiteralNode(
                        value=wdns.get_entity_abs_uri(vnode.qnode_id),
                        readable_label=self.get_qnode_label(vnode.qnode_id),
                        datatype=O.LiteralNodeDataType.Entity,
                        is_in_context=vnode.qnode_id == table.context.page_entity_id,
                    )
                    cpa_idmap[vnode.id] = sm.add_node(target)
                else:
                    target = sm.get_node(cpa_idmap[vnode.id])
            elif isinstance(vnode, CGLiteralValueNode):
                if vnode.id not in cpa_idmap:
                    target = O.LiteralNode(
                        value=vnode.value.to_string_repr(),
                        readable_label=vnode.label,
                        datatype=O.LiteralNodeDataType.String,
                    )
                    cpa_idmap[vnode.id] = sm.add_node(target)
                else:
                    target = sm.get_node(cpa_idmap[vnode.id])
            else:
                if vnode.id not in cpa_idmap:
                    # create a statement node
                    target = O.ClassNode(
                        abs_uri=WikidataNamespace.STATEMENT_URI,
                        rel_uri=wdns.get_rel_uri(WikidataNamespace.STATEMENT_URI),
                    )
                    cpa_idmap[vnode.id] = sm.add_node(target)
                else:
                    target = sm.get_node(cpa_idmap[vnode.id])

            sm.add_edge(
                O.Edge(
                    source=source.id,
                    target=target.id,
                    abs_uri=wdns.get_prop_abs_uri(edge.predicate),
                    rel_uri=wdns.get_prop_rel_uri(edge.predicate),
                    readable_label=self.get_pnode_label(edge.predicate),
                )
            )

        M.log("grams", semantic_model=sm, cpa=cpa, cta=cta)
        return sm

    def gen_equivalent_sm(
        self,
        sm: O.SemanticModel,
        strict: bool = True,
        force_inversion: bool = False,
        limited_invertible_props: Optional[Set[str]] = None,
        incorrect_invertible_props: Optional[Set[str]] = None,
    ):
        """Given a semantic model (not being modified), generate equivalent models by inferring inverse properties.

        Currently, we only inverse the properties, not qualifiers.

        Parameters
        ----------
        sm: the input semantic model (original)
        strict: whether to throw exception when target of an inverse property is not a class.
        force_inversion: only work when strict mode is set to false. Without force_inverse, we skip inverse properties,
                       otherwise, we generate an inverse model with a special class: wikibase:DummyClassForInversion
        limited_invertible_props: if provided, only generate inverse properties for these properties.
        incorrect_invertible_props: if provided, skip generating inverse properties for these properties.
        Returns
        -------
        """
        sm = self.norm_sm(sm)
        wdns = self.wdns

        if incorrect_invertible_props is None:
            incorrect_invertible_props = set()

        invertible_stmts: List[O.ClassNode] = []
        is_class_fn = lambda n1: isinstance(n1, O.ClassNode) or (
            isinstance(n1, O.LiteralNode) and wdns.is_abs_uri_qnode(n1.value)
        )

        for n in sm.iter_nodes():
            if isinstance(n, O.ClassNode) and wdns.is_abs_uri_statement(n.abs_uri):
                inedges = sm.in_edges(n.id)
                outedges = sm.out_edges(n.id)
                # only has one prop
                (prop,) = list({inedge.abs_uri for inedge in inedges})
                pid = wdns.get_prop_id(prop)
                stmt_has_value = False
                for outedge in outedges:
                    if outedge.abs_uri != prop:
                        # assert len(self.wdprops[self.get_prop_id(outedge.abs_uri)].inverse_properties) == 0, "Just to make sure" \
                        #                                                                                    "that qualifiers is not invertable. Otherwise, this algorithm will missing one generated SMs"
                        # character role has an inverse property: performer. They can be used as qualifier so nothing to do here just pass
                        pass
                    else:
                        stmt_has_value = True
                if (
                    len(self.wdprops[pid].inverse_properties) > 0
                    and pid not in incorrect_invertible_props
                    and (
                        limited_invertible_props is None
                        or pid in limited_invertible_props
                    )
                    and stmt_has_value
                ):
                    # invertible property
                    # people seem to misunderstand what inverse_property means in RDF;
                    # inverse doesn't apply to data property but only object property.
                    # so we catch the error here to detect what we should fix.
                    (outedge,) = [
                        outedge for outedge in outedges if outedge.abs_uri == prop
                    ]
                    targets_are_class = is_class_fn(sm.get_node(outedge.target))
                    if targets_are_class:
                        invertible_stmts.append(n)
                    elif strict:
                        raise Exception(f"{pid} is not invertible")
                    elif force_inversion:
                        assert isinstance(
                            sm.get_node(outedge.target), O.DataNode
                        ), "Clearly the model is wrong, you have an inverse property to a literal node"
                        invertible_stmts.append(n)

        # we have N statement, so we are going to have N! - 1 ways. It's literally a cartesian product
        all_choices = []
        for stmt in invertible_stmts:
            # assume that each statement only has one incoming link! fix the for loop if this assumption doesn't hold
            (inedge,) = sm.in_edges(stmt.id)
            choice: List[Tuple[O.ClassNode, Optional[str], Optional[str]]] = [
                (stmt, None, None)
            ]
            for invprop in self.wdprops[
                wdns.get_prop_id(inedge.abs_uri)
            ].inverse_properties:
                choice.append(
                    (
                        stmt,
                        wdns.get_prop_abs_uri(invprop),
                        wdns.get_prop_rel_uri(invprop),
                    )
                )
            all_choices.append(choice)

        n_choices = np.prod([len(c) for c in all_choices]) - 1
        if n_choices > 256:
            raise sm_metrics.PermutationExplosion("Too many possible semantic models")

        all_choices_perm: List[
            Tuple[Tuple[O.ClassNode, Optional[str], Optional[str]]]
        ] = list(itertools.product(*all_choices))
        assert all(
            invprop is None for _, invprop, _ in all_choices_perm[0]
        ), "First choice is always the current semantic model"
        new_sms = [sm]
        for choice_perm in all_choices_perm[1:]:
            new_sm = sm.copy()
            # we now change the statement from original prop to use the inverse prop (change direction)
            # if the invprop is not None
            for stmt, invprop_abs_uri, invprop_rel_uri in choice_perm:
                if invprop_abs_uri is None or invprop_rel_uri is None:
                    continue
                readable_label = self.get_pnode_label(wdns.get_prop_id(invprop_abs_uri))
                # assume that each statement only has one incoming link! fix the for loop if this assumption doesn't hold
                (inedge,) = sm.in_edges(stmt.id)
                # statement must have only one property
                (outedge,) = [
                    outedge
                    for outedge in sm.out_edges(stmt.id)
                    if outedge.abs_uri == inedge.abs_uri
                ]
                assert (
                    len(new_sm.get_edges_between_nodes(inedge.source, stmt.id)) == 1
                    and len(new_sm.get_edges_between_nodes(stmt.id, outedge.target))
                    == 1
                )
                new_sm.remove_edges_between_nodes(inedge.source, stmt.id)
                new_sm.remove_edges_between_nodes(stmt.id, outedge.target)

                target = sm.get_node(outedge.target)
                if not is_class_fn(target):
                    assert isinstance(target, O.DataNode)
                    dummy_class_node = O.ClassNode(
                        abs_uri=wdns.DUMMY_CLASS_FOR_INVERSION_URI,
                        rel_uri=wdns.get_rel_uri(wdns.DUMMY_CLASS_FOR_INVERSION_URI),
                    )
                    new_sm.add_node(dummy_class_node)
                    new_sm.add_edge(
                        O.Edge(
                            source=dummy_class_node.id,
                            target=target.id,
                            abs_uri=str(RDFS.label),
                            rel_uri="rdfs:label",
                        )
                    )
                    outedge_target = dummy_class_node.id
                else:
                    outedge_target = outedge.target
                new_sm.add_edge(
                    O.Edge(
                        source=outedge_target,
                        target=stmt.id,
                        abs_uri=invprop_abs_uri,
                        rel_uri=invprop_rel_uri,
                        approximation=outedge.approximation,
                        readable_label=readable_label,
                    )
                )
                new_sm.add_edge(
                    O.Edge(
                        source=stmt.id,
                        target=inedge.source,
                        abs_uri=invprop_abs_uri,
                        rel_uri=invprop_rel_uri,
                        approximation=inedge.approximation,
                        readable_label=readable_label,
                    )
                )
            new_sms.append(new_sm)
        return new_sms

    def get_entity_columns(self, sm: O.SemanticModel) -> List[int]:
        ent_columns = []
        for dnode in sm.iter_nodes():
            if isinstance(dnode, O.DataNode):
                inedges = sm.in_edges(dnode.id)
                if len(inedges) == 0:
                    continue
                assert len({edge.abs_uri for edge in inedges}) == 1, inedges
                edge_abs_uri = inedges[0].abs_uri
                if edge_abs_uri in self.ID_PROPS:
                    assert len(inedges) == 1, inedges
                    source = sm.get_node(inedges[0].source)
                    assert isinstance(
                        source, O.ClassNode
                    ) and not self.wdns.is_abs_uri_statement(source.abs_uri)
                    ent_columns.append(dnode.col_index)
        return ent_columns

    @classmethod
    def is_uri_column(cls, uri: str):
        """Test if an uri is for specifying the column"""
        return uri.startswith("http://example.com/table/")

    @staticmethod
    def get_column_uri(column_index: int):
        return f"http://example.com/table/{column_index}"

    @staticmethod
    def get_column_index(uri: str):
        assert WikidataSemanticModelHelper.is_uri_column(uri)
        return int(uri.replace("http://example.com/table/", ""))

    def extract_claims(
        self, tbl: LinkedTable, sm: O.SemanticModel, allow_multiple_ent: bool = True
    ):
        """Extract claims from the table given a semantic model.

        If an entity doesn't have link, its id will be null
        """
        # norm the semantic model first
        sm = self.norm_sm(sm)
        wdns = self.wdns
        schemas = {}
        for u in sm.iter_nodes():
            if not isinstance(u, O.ClassNode) or wdns.is_abs_uri_statement(u.abs_uri):
                continue

            schema = {"props": {}, "subject": None, "sm_node_id": u.id}
            for us_edge in sm.out_edges(u.id):
                if us_edge.abs_uri in self.ID_PROPS:
                    v = sm.get_node(us_edge.target)
                    assert isinstance(v, O.DataNode)
                    assert schema["subject"] is None
                    schema["subject"] = v.col_index
                    continue

                s = sm.get_node(us_edge.target)
                assert isinstance(s, O.ClassNode) and wdns.is_abs_uri_statement(
                    s.abs_uri
                )
                assert wdns.is_abs_uri_property(us_edge.abs_uri)

                pnode = wdns.get_prop_id(us_edge.abs_uri)
                if pnode not in schema["props"]:
                    schema["props"][pnode] = []

                stmt = {
                    "index": len(schema["props"][pnode]),
                    "value": None,
                    "qualifiers": [],
                }
                schema["props"][pnode].append(stmt)
                for sv_edge in sm.out_edges(s.id):
                    v = sm.get_node(sv_edge.target)

                    assert wdns.is_abs_uri_property(sv_edge.abs_uri)
                    if sv_edge.abs_uri == us_edge.abs_uri:
                        assert stmt["value"] is None, "only one property"
                        # this is property
                        if isinstance(v, O.ClassNode):
                            stmt["value"] = {"type": "classnode", "value": v.id}
                        elif isinstance(v, O.DataNode):
                            stmt["value"] = {"type": "datanode", "value": v.col_index}
                        else:
                            assert isinstance(v, O.LiteralNode)
                            stmt["value"] = {"type": "literalnode", "value": v.value}
                    else:
                        # this is qualifier
                        if isinstance(v, O.ClassNode):
                            stmt["qualifiers"].append(
                                {
                                    "type": "classnode",
                                    "pnode": wdns.get_prop_id(sv_edge.abs_uri),
                                    "value": v.id,
                                }
                            )
                        elif isinstance(v, O.DataNode):
                            stmt["qualifiers"].append(
                                {
                                    "type": "datanode",
                                    "pnode": wdns.get_prop_id(sv_edge.abs_uri),
                                    "value": v.col_index,
                                }
                            )
                        else:
                            assert isinstance(v, O.LiteralNode)
                            stmt["qualifiers"].append(
                                {
                                    "type": "literalnode",
                                    "pnode": wdns.get_prop_id(sv_edge.abs_uri),
                                    "value": v.value,
                                }
                            )
            schemas[u.id] = schema

        assert all(
            c.index == ci for ci, c in enumerate(tbl.table.columns)
        ), "Cannot handle table with missing columns yet"

        records = [{} for ri in range(len(tbl.table.columns[0].values))]
        node2ents = {}

        # extract data props first
        for cid, schema in schemas.items():
            ci = schema["subject"]
            col = tbl.table.columns[ci]
            for ri, val in enumerate(col.values):
                # get entities
                qnode_ids = sorted(
                    {
                        e.entity_id
                        for e in tbl.links[ri][ci]
                        if e.entity_id is not None and e.start < e.end
                    }
                )
                if len(qnode_ids) == 0:
                    # create new entity
                    ents = [
                        {
                            "id": f"{ri}-{ci}",
                            "column": ci,
                            "row": ri,
                            "uri": None,
                            "label": val,
                            "props": {},
                        }
                    ]
                else:
                    ents = [
                        {
                            "id": qnode_id,
                            "uri": wdns.get_entity_abs_uri(qnode_id),
                            "label": self.wdentities[qnode_id].label,
                            "props": {},
                        }
                        for qnode_id in qnode_ids
                    ]

                if len(ents) > 1:
                    if not allow_multiple_ent:
                        raise Exception("Encounter multiple entities")

                for prop, stmts in schema["props"].items():
                    for ent in ents:
                        assert prop not in ent["props"]
                        ent["props"][prop] = [
                            {"value": None, "qualifiers": {}} for stmt in stmts
                        ]
                        for stmt in stmts:
                            # set statement property
                            if stmt["value"]["type"] == "classnode":
                                # do it in later phase
                                pass
                            elif stmt["value"]["type"] == "datanode":
                                tci = stmt["value"]["value"]
                                ent["props"][prop][stmt["index"]][
                                    "value"
                                ] = tbl.table.columns[tci].values[ri]
                            else:
                                assert stmt["value"]["type"] == "literalnode"
                                ent["props"][prop][stmt["index"]]["value"] = stmt[
                                    "value"
                                ]["value"]

                            # set statement qualifiers
                            for qual in stmt["qualifiers"]:
                                if (
                                    qual["pnode"]
                                    not in ent["props"][prop][stmt["index"]][
                                        "qualifiers"
                                    ]
                                ):
                                    ent["props"][prop][stmt["index"]]["qualifiers"][
                                        qual["pnode"]
                                    ] = []
                                if qual["type"] == "classnode":
                                    # do it in later phase
                                    pass
                                elif qual["type"] == "datanode":
                                    tci = qual["value"]
                                    ent["props"][prop][stmt["index"]]["qualifiers"][
                                        qual["pnode"]
                                    ].append(tbl.table.columns[tci].values[ri])
                                elif qual["type"] == "literalnode":
                                    ent["props"][prop][stmt["index"]]["qualifiers"][
                                        qual["pnode"]
                                    ].append(qual["value"])

                for ent in ents:
                    assert (ent["id"], ci) not in records[ri]
                    records[ri][ent["id"], ci] = ent
                node2ents[schema["sm_node_id"], ri] = [ent for ent in ents]

        for cid, schema in schemas.items():
            ci = schema["subject"]
            col = tbl.table.columns[ci]
            for ri in range(len(col.values)):
                ulst = node2ents[schema["sm_node_id"], ri]
                for prop, stmts in schema["props"].items():
                    for stmt in stmts:
                        if stmt["value"]["type"] == "classnode":
                            vlst = node2ents[stmt["value"]["value"], ri]
                            for u in ulst:
                                assert len(vlst) > 0
                                u["props"][prop][stmt["index"]]["value"] = vlst[0]["id"]
                                if len(vlst) > 1:
                                    # this statement must not have other qualifiers, so that v can be a list
                                    # and we can create extra statement
                                    assert len(stmt["qualifiers"]) == 0
                                    for v in vlst[1:]:
                                        u["props"][prop].append(
                                            {"value": v["id"], "qualifiers": {}}
                                        )
                        for qual in stmt["qualifiers"]:
                            if qual["type"] == "classnode":
                                for u in ulst:
                                    vlst = node2ents[qual["value"], ri]
                                    u["props"][prop][stmt["index"]]["qualifiers"][
                                        qual["pnode"]
                                    ] = [v["id"] for v in vlst]

        new_records = []
        for ri, record in enumerate(records):
            new_record = {}
            for (ent_id, ci), ent in record.items():
                if ent_id in new_record:
                    # merge the entity
                    for pid, _stmts in ent["props"].items():
                        if pid not in new_record[ent_id]["props"]:
                            new_record[ent_id]["props"][pid] = _stmts
                        else:
                            for _stmt in _stmts:
                                if not any(
                                    x == _stmt for x in new_record[ent_id]["props"][pid]
                                ):
                                    new_record[ent_id]["props"][pid].append(_stmt)
                else:
                    new_record[ent_id] = ent
            new_records.append(new_record)
        return records

    def get_qnode_label(self, qid: str):
        """Get WDEntity label from id"""
        if qid in self.wdclasses:
            label = self.wdclasses[qid].label
        elif qid in self.wdentities:
            label = self.wdentities[qid].label
        else:
            label = self.wdentity_labels.get(qid, qid)
        return f"{label} ({qid})"

    def get_pnode_label(self, pid: str):
        """Get PNode label from id"""
        if pid not in self.wdprops:
            return pid
        return f"{self.wdprops[pid].label} ({pid})"



from typing import List, Mapping, Tuple, Callable, Dict, Union
from grams.algorithm.literal_matchers.text_parser import (
    ParsedTextRepr,
)
from kgdata.wikidata.models import WDValue, WDValueType, WDEntity
from grams.algorithm.literal_matchers.string_match import (
    string_exact_test,
    monolingual_exact_test,
)
from grams.algorithm.literal_matchers.number_match import quantity_test
from grams.algorithm.literal_matchers.coordinate_match import globecoordinate_test
from grams.algorithm.literal_matchers.datetime_match import time_test
from grams.algorithm.literal_matchers.entity_match import entity_similarity_test
from grams.algorithm.literal_matchers.types import LiteralMatchKit
from sm.misc.funcs import import_func


MatchFunc = Callable[[WDValue, ParsedTextRepr, LiteralMatchKit], Tuple[bool, float]]


class LiteralMatch:
    STRING = ".string_exact_test"
    QUANTITY = ".quantity_test"
    GLOBECOORDINATE = ".globecoordinate_test"
    TIME = ".time_test"
    MONOLINGUAL_TEXT = ".monolingual_exact_test"
    ENTITY = ""

    def __init__(self, wdentities: Mapping[str, WDEntity]):
        self.match_kit = LiteralMatchKit(wdentities)

        self.type2func: Dict[
            WDValueType,
            List[MatchFunc],
        ] = {}

        self.type2func["string"] = self.cfg2funcs(LiteralMatch.STRING)
        self.type2func["quantity"] = self.cfg2funcs(LiteralMatch.QUANTITY)
        self.type2func["globecoordinate"] = self.cfg2funcs(LiteralMatch.GLOBECOORDINATE)
        self.type2func["time"] = self.cfg2funcs(LiteralMatch.TIME)
        self.type2func["monolingualtext"] = self.cfg2funcs(
            LiteralMatch.MONOLINGUAL_TEXT
        )
        self.type2func["wikibase-entityid"] = self.cfg2funcs(LiteralMatch.ENTITY)

    def match(
        self, pval: WDValue, val: ParsedTextRepr, skip_unmatch: bool = True
    ) -> List[Tuple[MatchFunc, Tuple[bool, float]]]:
        if skip_unmatch:
            return [
                (func, m)
                for func in self.type2func[pval.type]
                if (m := func(pval, val, self.match_kit))[0]
            ]
        return [
            (func, func(pval, val, self.match_kit))
            for func in self.type2func[pval.type]
        ]

    @staticmethod
    def cfg2funcs(cfg: str) -> List[MatchFunc]:
        funcs = []
        if len(cfg) == 0:
            return funcs
        for path in cfg.split(","):
            if path[0] == ".":
                # relative import
                funcs.append(
                    {
                        "string_exact_test": string_exact_test,
                        "monolingual_exact_test": monolingual_exact_test,
                        "quantity_test": quantity_test,
                        "globecoordinate_test": globecoordinate_test,
                        "time_test": time_test,
                        "entity_similarity_test": entity_similarity_test,
                    }[path[1:]]
                )
            else:
                funcs.append(import_func(path))
        return funcs
